from uplogic.nodes import ULOutSocket
from uplogic.nodes import ULParameterNode
from uplogic.utils import STATUS_WAITING
from uplogic.utils import is_waiting
from uplogic.utils import is_invalid


class ULFormattedString(ULParameterNode):
    def __init__(self):
        ULParameterNode.__init__(self)
        self.format_string = None
        self.value_a = None
        self.value_b = None
        self.value_c = None
        self.value_d = None
        self.OUT = ULOutSocket(self, self.get_out)

    def get_out(self):
        socket = self.get_output('out')
        if socket is None:
            format_string = self.get_input(self.format_string)
            value_a = self.get_input(self.value_a)
            value_b = self.get_input(self.value_b)
            value_c = self.get_input(self.value_c)
            value_d = self.get_input(self.value_d)
            if is_invalid(format_string):
                return STATUS_WAITING
            if is_waiting(value_a, value_b, value_c, value_d):
                return STATUS_WAITING
            result = format_string.format(value_a, value_b, value_c, value_d)
            return self.set_output('out', result)
        return socket

    def evaluate(self):
        self._set_ready()



from uplogic.nodes import ULParameterNode
from uplogic.nodes import ULOutSocket
from uplogic.utils import STATUS_INVALID, STATUS_WAITING
from uplogic.utils import is_waiting


class ULGetPyInstanceAttr(ULParameterNode):
    def __init__(self):
        ULParameterNode.__init__(self)
        self.instance = None
        self.attr = None
        self.OUT = ULOutSocket(self, self.get_done)

    def get_done(self):
        instance = self.get_input(self.instance)
        attr = self.get_input(self.attr)
        if is_waiting(instance, attr):
            return STATUS_WAITING
        return getattr(instance, attr, STATUS_INVALID)

    def evaluate(self):
        self._set_ready()



from uplogic.nodes import ULConditionNode
from uplogic.nodes import ULOutSocket


class ULTrueFalse(ULConditionNode):
    def __init__(self):
        ULConditionNode.__init__(self)
        self.state = None
        self.TRUE = ULOutSocket(self, self.get_true_value)
        self.FALSE = ULOutSocket(self, self.get_false_value)

    def get_true_value(self):
        return self.get_input(self.state)

    def get_false_value(self):
        return not self.get_input(self.state)

    def evaluate(self):
        self._set_ready()



from uplogic.nodes import ULActionNode
from uplogic.nodes import ULOutSocket
from uplogic.utils import is_waiting
from uplogic.utils import not_met
from uplogic.shaders import FXAA, HBAO, SSAO, Vignette, Brightness, ChromaticAberration, Grayscale, Levels, Mist


class ULAddFilter(ULActionNode):
    def __init__(self):
        ULActionNode.__init__(self)
        self.condition = None
        self.pass_idx = None
        self.brightness = None
        self.power = None
        self.color = None
        self.start = None
        self.end = None
        self.filter = None
        self.filter_type = 'FXAA'
        self.OUT = ULOutSocket(self, self._get_done)

    def _get_done(self):
        return self.done

    def evaluate(self):
        self.done = False
        condition = self.get_input(self.condition)
        ftype = self.filter_type
        if not_met(condition) or self.filter:
            if self.filter:
                if ftype == 'VIGNETTE':
                    self.filter.settings['power'] = self.get_input(self.power)
                    self.filter.settings['color'] = self.get_input(self.color)
                elif ftype == 'BRIGHTNESS':
                    self.filter.settings['brightness'] = self.get_input(self.brightness)
                elif ftype == 'CRHOMAB':
                    self.filter.settings['power'] = self.get_input(self.power)
                elif ftype == 'GRAYSCALE':
                    self.filter.settings['power'] = self.get_input(self.power)
                elif ftype == 'LEVELS':
                    self.filter.settings['color'] = self.get_input(self.color)
                elif ftype == 'MIST':
                    self.filter.settings['power'] = self.get_input(self.power)
                    self.filter.settings['color'] = self.get_input(self.color)
                    self.filter.settings['start'] = self.get_input(self.start)
                    self.filter.settings['end'] = self.get_input(self.end)
            return
        self._set_ready()

        if ftype == 'FXAA':
            self.filter = FXAA(self.get_input(self.pass_idx))
        elif ftype == 'HBAO':
            self.filter = HBAO(self.get_input(self.pass_idx))
        elif ftype == 'SSAO':
            self.filter = SSAO(self.get_input(self.pass_idx))
        elif ftype == 'VIGNETTE':
            self.filter = Vignette(
                self.get_input(self.power),
                self.get_input(self.color),
                self.get_input(self.pass_idx)
            )
        elif ftype == 'BRIGHTNESS':
            self.filter = Brightness(
                self.get_input(self.brightness),
                self.get_input(self.pass_idx)
            )
        elif ftype == 'CRHOMAB':
            self.filter = ChromaticAberration(
                self.get_input(self.power),
                self.get_input(self.pass_idx)
            )
        elif ftype == 'GRAYSCALE':
            self.filter = Grayscale(
                self.get_input(self.power),
                self.get_input(self.pass_idx)
            )
        elif ftype == 'LEVELS':
            self.filter = Levels(
                self.get_input(self.color),
                self.get_input(self.pass_idx)
            )
        elif ftype == 'MIST':
            self.filter = Mist(
                self.get_input(self.start),
                self.get_input(self.end),
                self.get_input(self.color),
                self.get_input(self.power),
                self.get_input(self.pass_idx)
            )
        self.done = True



from uplogic.nodes import ULOutSocket
from uplogic.nodes import ULParameterNode
from uplogic.utils import STATUS_WAITING
from uplogic.utils import is_invalid


class ULGetActuatorValue(ULParameterNode):

    def __init__(self):
        ULParameterNode.__init__(self)
        self.game_obj = None
        self.act_name = None
        self.field = None
        self.OUT = ULOutSocket(self, self.get_actuator)

    def get_actuator(self):
        socket = self.get_output('actuator')
        if socket is None:
            game_obj = self.get_input(self.game_obj)
            act_name = self.get_input(self.act_name)
            field = self.get_input(self.field)
            if is_invalid(game_obj, act_name, field):
                return STATUS_WAITING
            if act_name not in game_obj.actuators:
                return STATUS_WAITING
            return self.set_output(
                'actuator',
                getattr(game_obj.actuators[act_name], field)
            )
        return socket

    def evaluate(self):
        self._set_ready()



from uplogic.nodes import ULConditionNode
from uplogic.utils import is_waiting
from uplogic.utils import not_met


class ULTimer(ULConditionNode):

    def __init__(self):
        ULConditionNode.__init__(self)
        self.condition = None
        self.delta_time = None
        self._trigger = -1
        self.network = None

    def setup(self, network):
        self.network = network

    def evaluate(self):
        condition = self.get_input(self.condition)
        delta_time = self.get_input(self.delta_time)
        if is_waiting(delta_time):
            return
        self._set_ready()
        now = self.network.timeline

        if not not_met(condition):
            self._trigger = now + delta_time

        if self._trigger == -1 or now < self._trigger:
            self._set_value(False)
        else:
            self._set_value(True)
            self._trigger = -1



from .buoyancy import ULFloatsam  # noqa
from .buoyancy import ULShip  # noqa
from .collision import on_collision  # noqa
from .vehicle import FOURWD  # noqa
from .vehicle import FWD  # noqa
from .vehicle import RWD  # noqa
from .vehicle import ULVehicle  # noqa
from .character import ULCharacter  # noqa
from .constraints import create_constraint  # noqa
from .constraints import remove_constraint  # noqa
from .constraints import ULSpring
from .constraints import ULTrackTo



from .gamepad import LS  # noqa
from .gamepad import RS  # noqa
from .gamepad import SONY  # noqa
from .gamepad import XBOX  # noqa
from .gamepad import gamepad_down  # noqa
from .gamepad import gamepad_stick  # noqa
from .gamepad import gamepad_trigger  # noqa
from .gamepad import gamepad_tap  # noqa
from .gamepad import gamepad_up  # noqa
from .gamepad import ULGamePad  # noqa
from .gamepad import ULGamepadLook  # noqa
from .keyboard import KEYBOARD_EVENTS  # noqa
from .keyboard import key_down  # noqa
from .keyboard import key_pulse  # noqa
from .keyboard import key_tap  # noqa
from .keyboard import key_up  # noqa
from .mouse import LMB  # noqa
from .mouse import MMB  # noqa
from .mouse import MOUSE_EVENTS  # noqa
from .mouse import RMB  # noqa
from .mouse import get_mouse_position  # noqa
from .mouse import mouse_down  # noqa
from .mouse import mouse_moved  # noqa
from .mouse import mouse_tap  # noqa
from .mouse import mouse_up  # noqa
from .mouse import mouse_wheel  # noqa
from .mouse import set_mouse_position  # noqa
from .mouse import ULMouse  # noqa
from .mouse import ULMouseLook  # noqa
from .vr import get_vr_headset_data  # noqa
from .vr import ULControllerVR  # noqa
from .vr import ULHeadsetVR  # noqa
from .vr import ULCharacterVR  # noqa


from bge import render
from mathutils import Vector
from uplogic.nodes import ULOutSocket
from uplogic.nodes import ULParameterNode


class ULGetResolution(ULParameterNode):
    def __init__(self):
        ULParameterNode.__init__(self)
        self.width = None
        self.height = None
        self.res = None
        self.WIDTH = ULOutSocket(self, self.get_width)
        self.HEIGHT = ULOutSocket(self, self.get_height)
        self.RES = ULOutSocket(self, self.get_res)

    def get_width(self):
        return render.getWindowWidth()

    def get_height(self):
        return render.getWindowHeight()

    def get_res(self):
        return Vector((self.width, self.height))

    def evaluate(self):
        self._set_ready()



from uplogic.nodes import ULOutSocket
from uplogic.nodes import ULParameterNode
from uplogic.utils import STATUS_WAITING
from uplogic.utils import is_invalid
from bge.types import KX_GameObject as GameObject


class ULGetSensorValue(ULParameterNode):

    def __init__(self):
        ULParameterNode.__init__(self)
        self.game_obj = None
        self.sens_name = None
        self.field = None
        self.OUT = ULOutSocket(self, self.get_sensor)

    def get_sensor(self):
        game_obj: GameObject = self.get_input(self.game_obj)
        sens_name: str = self.get_input(self.sens_name)
        field = self.get_input(self.field)
        if is_invalid(game_obj, sens_name, field):
            return STATUS_WAITING
        if sens_name not in game_obj.sensors:
            return STATUS_WAITING
        return getattr(game_obj.sensors[sens_name], field)

    def evaluate(self):
        self._set_ready()



from uplogic.nodes import ULOutSocket
from uplogic.nodes import ULParameterNode
from uplogic.utils import STATUS_WAITING
from uplogic.utils import is_invalid


class ULListDuplicate(ULParameterNode):
    def __init__(self):
        ULParameterNode.__init__(self)
        self.condition = None
        self.items = None
        self.OUT = ULOutSocket(self, self.get_points)

    def get_points(self):
        list_d = self.get_input(self.items)
        if is_invalid(list_d):
            return STATUS_WAITING
        return list_d.copy()

    def evaluate(self):
        self._set_ready()



from .shader import ULFilter
from .fxaa import FXAA
from .brightness import Brightness
from .vignette import Vignette
from .grayscale import Grayscale
from .hbao import HBAO
from .ssao import SSAO
from .mist import Mist
from .levels import Levels
from .chromaticaberration import ChromaticAberration
from .shader import load_glsl


from bge import logic
from bge import render
from bgui.bgui_utils import Layout
from bgui.bgui_utils import System
from bgui.label import Label
from io import StringIO
from uplogic.data import GlobalDB
from uplogic.utils import lerp
from uplogic.events import schedule_callback
import sys


def set_depth(depth):
    LoggerLayout.max_msg = depth


def enable():
    sys.stdout = Logger()
    sys.stderr = ErrorLogger()


def disable():
    sys.stdout = sys.__stdout__
    sys.stderr = sys.__stderr__


class Logger(StringIO):

    def write(self, __s: str) -> int:
        log(__s)
        sys.__stdout__.write(__s)


class ErrorLogger(StringIO):
    def write(self, __s: str) -> int:
        error(__s)
        sys.__stderr__.write(__s)


class LoggerLayout(Layout):
    colors = {
        'INFO': [1, 1, 1, 1],
        'WARNING': [1, 1, .3, 1],
        'ERROR': [1, .3, .3, 1]
    }
    max_msg = 20
    opacity = 1

    def __init__(self, system, data):
        super().__init__(system, data)
        self.messages: list[Label] = []
        self.fade_event = None
        if disable not in logic.getCurrentScene().onRemove:
            logic.getCurrentScene().onRemove.append(disable)

    def fade_out(self):
        for msg in self.messages:
            msg.color[3] = lerp(msg.color[3], self.messages[0].color[3], .02)
        if self.messages[-1].color[3] != .1:
            self.fade_event = schedule_callback(self.fade_out)
        else:
            self.fade_event = None

    def add_message(self, msg, type='INFO'):
        if len(self.messages) > self.max_msg:
            self._remove_widget(self.messages[0])
            self.messages.pop(0)
        if self._system:
            self.messages.append(Label(
                self,
                text=str(msg),
                pt_size=0,
                color=self.colors.get(type, [1, 1, 1, 1]).copy(),
                pos=[0.02, 0],
                outline_color=(1, 0, 0, 1)
            ))
        for i, log in enumerate(reversed(self.messages)):
            size = 11 * (1080 / render.getWindowHeight())
            log.position[1] = 13 * (i+1)
            log.color[3] = 1 - (i/(self.max_msg + 5))
            log.pt_size = size
        if self.fade_event:
            self.fade_event.cancel()
        self.fade_event = schedule_callback(self.fade_out, 3)


def get_logger() -> LoggerLayout:
    loggers = GlobalDB.retrieve('uplogic.loggers')
    layout = loggers.get('default')
    if layout is None:
        system = System()
        system.load_layout(LoggerLayout, None)
        loggers.put('default', system.layout)
        layout = system.layout
        logic.getCurrentScene().post_draw.append(system.run)
    return layout


def log(msg):
    logger = get_logger()
    if logger is None:
        return
    for msg in str(msg).split('\n'):
        if msg:
            msg = msg.replace('  ', '    ')
            logger.add_message(f'INFO:\t{msg}')


def warning(msg):
    logger = get_logger()
    if logger is None:
        return
    for msg in str(msg).split('\n'):
        if msg:
            msg.replace('  ', '    ')
            logger.add_message(f'WARNING:\t{msg}', 'WARNING')


def error(msg):
    logger = get_logger()
    if logger is None:
        return
    for msg in str(msg).split('\n'):
        if msg:
            msg.replace('  ', '    ')
            logger.add_message(f'ERROR:\t{msg}', 'ERROR')



from uplogic.events import send
from uplogic.nodes import ULActionNode
from uplogic.nodes import ULOutSocket
from uplogic.utils import STATUS_INVALID
from uplogic.utils import is_invalid
from uplogic.utils import is_waiting
from uplogic.utils import not_met


class ULDispatchEvent(ULActionNode):
    def __init__(self):
        ULActionNode.__init__(self)
        self.condition = None
        self.subject = None
        self.body = None
        self.target = None
        self.old_subject = None
        self.done = None
        self.OUT = ULOutSocket(self, self.get_done)

    def get_done(self):
        return self.done

    def evaluate(self):
        self.done = False
        condition = self.get_input(self.condition)
        if not_met(condition):
            self._set_ready()
            return
        subject = self.get_input(self.subject)
        body = self.get_input(self.body)
        if body is STATUS_INVALID:
            body = None
        target = self.get_input(self.target)
        if is_waiting(body, target):
            return
        if is_invalid(subject):
            return
        self._set_ready()
        send(subject, body, target)
        self.done = True



from uplogic.nodes import ULOutSocket
from uplogic.nodes import ULParameterNode
from uplogic.utils import STATUS_WAITING
from uplogic.utils import is_invalid


class ULGetCollection(ULParameterNode):
    def __init__(self):
        ULParameterNode.__init__(self)
        self.collection = None
        self.OUT = ULOutSocket(self, self.get_collection)

    def get_collection(self):
        collection = self.get_input(self.collection)
        if is_invalid(collection):
            return STATUS_WAITING
        return collection

    def evaluate(self):
        self._set_ready()



import bpy
from mathutils import Quaternion
from mathutils import Matrix
from mathutils import Vector
from bge import logic
from bge.types import KX_GameObject as GameObject
from uplogic.utils.errors import NoXRSessionError


def get_vr_headset_data() -> tuple[Vector, Matrix]:
    """Get the current position and orientation of connected VR headset.

    :returns: `tuple` of (position: `Vector`, orientation: `Matrix`)
    """
    session = bpy.context.window_manager.xr_session_state
    if not session:
        raise NoXRSessionError
    return (Vector(session.viewer_pose_location), Quaternion(session.viewer_pose_rotation).to_matrix())


class ULControllerVR():
    """Wrapper class for VR Controllers.

    This wrapper provides `position`, `position_aim`, `orientation`,
    `orientation_aim`, `trigger` and `thumbstick` attributes.

    :param `idx`: Controller index; 0 for left, 1 for right controller.
    """

    def __init__(self, idx: int=0) -> None:
        self.idx = idx
        self.stick_threshold = 0.0
        self.session = bpy.context.window_manager.xr_session_state
        if not self.session:
            raise NoXRSessionError

    @property
    def position(self) -> Vector:
        """The global position of the controller."""
        return Vector(self.session.controller_grip_location_get(bpy.context, self.idx))

    @position.setter
    def position(self, val):
        print("Attribute 'position' of 'ULControllerVR' is read-only!")

    @property
    def orientation(self) -> Matrix:
        """The global orientation of the controller."""
        return Quaternion(self.session.controller_grip_rotation_get(bpy.context, self.idx)).to_matrix()

    @orientation.setter
    def orientation(self, val):
        print("Attribute 'orientation' of 'ULControllerVR' is read-only!")

    @property
    def position_aim(self) -> Vector:
        """The global position of the tip of the controller."""
        return Vector(self.session.controller_aim_location_get(bpy.context, self.idx))

    @position_aim.setter
    def position_aim(self, val):
        print("Attribute 'position_aim' of 'ULControllerVR' is read-only!")

    @property
    def orientation_aim(self) -> Matrix:
        """The global orientation of the tip of the controller."""
        return Quaternion(self.session.controller_aim_rotation_get(bpy.context, self.idx)).to_matrix()

    @orientation_aim.setter
    def orientation_aim(self, val):
        print("Attribute 'orientation_aim' of 'ULControllerVR' is read-only!")

    @property
    def aim(self) -> Vector:
        """Targeting vector of the controller."""
        aim = self.position_aim - self.position
        return aim.normalized()

    @aim.setter
    def aim(self, val):
        print("Attribute 'aim' of 'ULControllerVR' is read-only!")

    @property
    def trigger(self) -> float:
        """The intensity with which the trigger on the controller is pressed."""
        return self.session.action_state_get(
            bpy.context,
            'blender_default',
            'teleport',
            f"/user/hand/{'left' if self.idx == 0 else 'right'}"
        )[0]

    @trigger.setter
    def trigger(self, val):
        print("Attribute 'trigger' of 'ULControllerVR' is read-only!")

    @property
    def thumbstick(self) -> Vector:
        """Stick values for the controller."""
        if self.idx == 0:
            x = self.session.action_state_get(
                bpy.context,
                'blender_default',
                'fly_right',
                '/user/hand/left'
            )[0]
            y = self.session.action_state_get(
                bpy.context,
                'blender_default',
                'fly_forward',
                '/user/hand/left'
            )[0]
            return Vector((x, y))
        else:
            x = self.session.action_state_get(
                bpy.context,
                'blender_default',
                'fly_turnright',
                '/user/hand/right'
            )[0]
            y = self.session.action_state_get(
                bpy.context,
                'blender_default',
                'fly_up',
                '/user/hand/right'
            )[0]
            return Vector((x, y))

    @thumbstick.setter
    def thumbstick(self, val):
        print("Attribute 'thumbstick' of 'ULControllerVR' is read-only!")


class ULHeadsetVR():
    """Wrapper class for a VR Headset.

    This wrapper provides a `position` and an `orientation`
    attribute.
    """
    def __init__(self) -> None:
        self.session = bpy.context.window_manager.xr_session_state
        if not self.session:
            raise NoXRSessionError

    @property
    def position(self) -> Vector:
        return Vector(self.session.viewer_pose_location)

    @position.setter
    def position(self, val):
        print("Attribute 'position' of 'ULHeadsetVR' is read-only!")

    @property
    def orientation(self) -> Matrix:
        return Quaternion(self.session.viewer_pose_rotation).to_matrix()

    @orientation.setter
    def orientation(self, val):
        print("Attribute 'orientation' of 'ULHeadsetVR' is read-only!")


class ULHeadsetVRWrapper(ULHeadsetVR):
    """Wrapper class for a VR Headset to be used for audio calculations.

    Not intended for manual use.
    """
    @property
    def worldPosition(self) -> Vector:
        return Vector(self.session.viewer_pose_location)

    @worldPosition.setter
    def worldPosition(self, val):
        print("Attribute 'worldPosition' of 'ULHeadsetVRWrapper' is read-only!")

    @property
    def worldOrientation(self) -> Matrix:
        return Quaternion(self.session.viewer_pose_rotation).to_matrix()

    @worldOrientation.setter
    def worldOrientation(self, val):
        print("Attribute 'worldOrientation' of 'ULHeadsetVRWrapper' is read-only!")
    
    def rayCast(
        self,
        obj_to,
        obj_from,
        distance,
        xray
    ):
        return logic.getCurrentScene().active_camera.rayCast(obj_to, obj_from, distance, xray=xray)


class ULCharacterVR():
    """Wrapper class for all VR Devices. This wrapper contains 2 `ULControllerVR` objects as well as one
    `ULHeadsetVR` object.

    Optionally, `KX_GameObjects` can be defined for both left and right controller. These objects will by
    automatically synched with their respective controller's position and orientation.

    This class provides a `position` an `orientation` and a `scale` attribute.
    """
    def __init__(
        self,
        left_hand_object: GameObject = None,
        right_hand_object: GameObject = None
    ) -> None:
        self.session = bpy.context.window_manager.xr_session_state
        if not self.session:
            raise NoXRSessionError
        self.hand_left = ULControllerVR(0)
        self.hand_right = ULControllerVR(1)
        self.head = ULHeadsetVR()
        self.hand_left_object = left_hand_object
        self.hand_right_object = right_hand_object
        if left_hand_object is not None or right_hand_object is not None:
            logic.getCurrentScene().pre_draw.append(self.update)
    
    def update(self):
        if self.hand_left_object:
            self.hand_left_object.worldPosition = self.hand_left.position_aim
            self.hand_left_object.worldOrientation = self.hand_left.orientation_aim
        if self.hand_right_object:
            self.hand_right_object.worldPosition = self.hand_right.position_aim
            self.hand_right_object.worldOrientation = self.hand_right.orientation_aim

    @property
    def position(self) -> Vector:
        return Vector(self.session.navigation_location)

    @position.setter
    def position(self, val):
        print("Attribute 'position' of 'ULHeadsetVR' is read-only!")

    @property
    def orientation(self) -> Matrix:
        return Quaternion(self.session.navigation_rotation).to_matrix()

    @orientation.setter
    def orientation(self, val):
        print("Attribute 'orientation' of 'ULHeadsetVR' is read-only!")

    @property
    def scale(self) -> float:
        return self.session.navigation_scale

    @scale.setter
    def scale(self, val):
        print("Attribute 'scale' of 'ULHeadsetVR' is read-only!")



from uplogic.nodes import ULOutSocket
from uplogic.nodes import ULParameterNode
from uplogic.utils import STATUS_WAITING
from uplogic.utils import is_invalid
from uplogic.utils import is_waiting


class ULListIndex(ULParameterNode):
    def __init__(self):
        ULParameterNode.__init__(self)
        self.items = None
        self.index = None
        self.OUT = ULOutSocket(self, self.get_val)

    def get_val(self):
        list_d = self.get_input(self.items)
        index = self.get_input(self.index)
        if is_invalid(list_d):
            return STATUS_WAITING
        if is_waiting(index):
            return STATUS_WAITING
        if index <= len(list_d) - 1:
            return list_d[index]
        return STATUS_WAITING

    def evaluate(self):
        self._set_ready()



from uplogic.nodes import ULConditionNode


class ULOnNextTick(ULConditionNode):

    def __init__(self):
        ULConditionNode.__init__(self)
        self.input_condition = None
        self._activated = 0

    def evaluate(self):
        input_condition = self.get_input(self.input_condition)
        self._set_ready()
        if self._activated == 1:
            self._set_value(True)
            if not input_condition:
                self._activated = 0
        elif input_condition:
            self._set_value(False)
            self._activated = 1
        elif self._activated == 0:
            self._set_value(False)



from bge import logic
from uplogic.data import GlobalDB


class ULActionSystem():
    '''System for managing actions started using `ULAction`. This class is
    usually addressed indirectly through `ULAction` and is not intended for
    manual use.
    '''
    layers: dict = {}

    def __init__(self, name: str):
        self.actions: list = []
        self.scene = scene = logic.getCurrentScene()
        self.listener = scene.active_camera
        self.old_lis_pos = self.listener.worldPosition.copy()
        GlobalDB.retrieve('uplogic.animation').put(name, self)
        scene.pre_draw.append(self.update)
        scene.onRemove.append(self.shutdown)

    @classmethod
    def lock_layer(cls, action):
        """Lock a layer according to a `ULAction`.

        :param `action`: The `ULAction` whose layer will be locked.
        """
        layers = cls.layers.get(action.game_object, {})
        layers[action.layer] = action
        cls.layers[action.game_object] = layers

    @classmethod
    def free_layer(cls, action):
        """Allow for the layer of the given action to be used again.

        :param `action`: The `ULAction` whose layer will be freed.
        """
        layers = cls.layers.get(action.game_object, {})
        layers.pop(action.layer, None)
        cls.layers[action.game_object] = layers

    @classmethod
    def find_free_layer(cls, action):
        """Incrementally find the next free layer for a `ULAction`.

        :param `action`: The `ULAction` for which to find a free layer.
        """
        layers = cls.layers.get(action.game_object, {})
        action.layer = 0
        while action.layer in layers.keys():
            action.layer += 1

    @classmethod
    def check_layer(cls, action):
        """Check if the layer for this `ULAction` is free.

        :param `action`: The `ULAction` whose layer to check.

        :return: `True` if the layer is occupied, `False` if not
        """
        layers = cls.layers.get(action.game_object, {})
        return action.layer in layers.keys()
    
    @classmethod
    def get_layer(cls, game_object, layer=0):
        """Get the `ULAction` of an object on the given layer.

        :param `game_object`: The `KX_GameObject` on which the action is
        playing.
        :param `layer`: The layer on which the action is playing.

        :return: `ULAction` if layer is occupied, else `None`
        """
        return cls.layers.get(game_object, {}).get(layer, None)

    def update(self):
        """This is called each frame.
        """
        for action in self.actions:
            action.update()

    def add(self, action):
        '''Add a `ULAction` to this system.

        :param `action`: `ULAction` to add.
        '''
        self.actions.append(action)
        ULActionSystem.lock_layer(action)

    def remove(self, action):
        '''Remove a `ULAction` from this system.

        :param `action`: `ULAction` which to remove.
        '''
        action.stop()
        if action in self.actions:
            self.actions.remove(action)
        ULActionSystem.free_layer(action)

    def shutdown(self):
        '''Shutdown and remove this action system. This will stop all actions
        playing in this system.'''
        self.scene.pre_draw.remove(self.update)
        for action in self.actions.copy():
            self.remove(action)
        GlobalDB.retrieve('uplogic.animation').remove(self)


def get_action_system(system_name: str = 'default') -> ULActionSystem:
    """Get or create a `ULActionSystem` with the given name. Using more than one
    action system is highly discouraged.

    :param `system_name`: Look for this name.

    :returns: `ULActionSystem`, new system is created if none is found.
    """
    act_systems = GlobalDB.retrieve('uplogic.animation')
    if act_systems.check(system_name):
        return act_systems.get(system_name)
    else:
        return ULActionSystem(system_name)


from uplogic.nodes import ULParameterNode
from uplogic.nodes import ULOutSocket
from uplogic.utils import STATUS_WAITING
from uplogic.utils import is_waiting
from uplogic.utils import is_invalid


class ULClamp(ULParameterNode):

    def __init__(self):
        ULParameterNode.__init__(self)
        self.value = None
        self.range = None
        self.OUT = ULOutSocket(self, self.get_done)

    def get_done(self):
        socket = self.get_output('result')
        if socket is None:
            value = self.get_input(self.value)
            range_ft = self.get_input(self.range)
            if is_waiting(range_ft):
                return STATUS_WAITING
            if is_invalid(value):
                return STATUS_WAITING
            if range_ft.x == range_ft.y:
                return value
            if value < range_ft.x:
                value = range_ft.x
            if value > range_ft.y:
                value = range_ft.y
            return self.set_output('result', value)
        return socket

    def evaluate(self):
        self._set_ready()



from mathutils import Vector
from uplogic.nodes import ULParameterNode
from uplogic.nodes import ULOutSocket
from uplogic.utils import STATUS_WAITING
from uplogic.utils import is_invalid


class ULMath(ULParameterNode):

    def __init__(self):
        ULParameterNode.__init__(self)
        self.operand_a = None
        self.operand_b = None
        self.operator = None
        self.OUT = ULOutSocket(self, self.get_done)

    def get_done(self):
        socket = self.get_output('done')
        if socket is None:
            a = self.get_input(self.operand_a)
            b = self.get_input(self.operand_b)
            if is_invalid(a, b):
                return STATUS_WAITING
            if (a is None) or (b is None):
                return STATUS_WAITING
            else:
                if (
                    isinstance(a, Vector) and
                    isinstance(b, Vector)
                ):
                    return self.get_vec_vec_calc(a, b)
                elif isinstance(a, Vector):
                    return self.get_vec_calc(a, b)
                elif isinstance(b, Vector):
                    return self.get_vec_calc(b, a)
                return self.set_output('done', self.operator(a, b))
        return socket

    def evaluate(self):
        self._set_ready()

    def get_vec_calc(self, vec, num):
        if len(vec) == 4:
            return Vector(
                (
                    self.operator(vec.x, num),
                    self.operator(vec.y, num),
                    self.operator(vec.z, num),
                    self.operator(vec.w, num)
                )
            )
        else:
            return Vector(
                (
                    self.operator(vec.x, num),
                    self.operator(vec.y, num),
                    self.operator(vec.z, num)
                )
            )

    def get_vec_vec_calc(self, vec, vec2):
        if len(vec) == 4 and len(vec2) == 4:
            return Vector(
                (
                    self.operator(vec.x, vec2.x),
                    self.operator(vec.y, vec2.y),
                    self.operator(vec.z, vec2.z),
                    self.operator(vec.w, vec2.w)
                )
            )
        else:
            return Vector(
                (
                    self.operator(vec.x, vec2.x),
                    self.operator(vec.y, vec2.y),
                    self.operator(vec.z, vec2.z)
                )
            )



from mathutils import Vector
from uplogic.nodes import ULConditionNode
from uplogic.nodes import ULOutSocket
from uplogic.utils import LOGIC_OPERATORS
from uplogic.utils import STATUS_WAITING
from uplogic.utils import is_waiting


class ULCompareVectors(ULConditionNode):
    def __init__(self, operator='GREATER'):
        ULConditionNode.__init__(self)
        self.operator = operator
        self.all = None
        self.threshold = None
        self.param_a = None
        self.param_b = None
        self.OUT = ULOutSocket(self, self.get_result)

    def get_result(self):
        socket = self.get_output('result')
        if socket is None:
            a = self.get_input(self.param_a)
            b = self.get_input(self.param_b)
            all_values = self.get_input(self.all)
            operator = self.get_input(self.operator)
            threshold = self.get_input(self.threshold)
            if is_waiting(a, b, all_values, operator, threshold):
                return STATUS_WAITING
            if (
                not isinstance(a, Vector)
                or not isinstance(b, Vector)
            ):
                return STATUS_WAITING
            if operator > 1:  # eq and neq are valid for None
                if a is None:
                    return STATUS_WAITING
                if b is None:
                    return STATUS_WAITING
            if operator is None:
                return STATUS_WAITING
            return self.set_output(
                'result',
                self.get_vec_val(
                    operator,
                    a,
                    b,
                    all_values,
                    threshold
                )
            )
        return socket

    def get_vec_val(self, op, a, b, xyz, threshold):
        for ax in ['x', 'y', 'z']:
            av = getattr(a, ax)
            bv = getattr(b, ax)
            av = bv if abs(av - bv) < threshold else av
            if xyz[ax] and not LOGIC_OPERATORS[op](av, bv):
                return False
        return True

    def evaluate(self):
        self._set_ready()



from .shader import ULFilter


glsl = """
//----------------------------------------------------------------------------------
//
// Copyright (c) 2014, NVIDIA CORPORATION. All rights reserved.
//
// Redistribution and use in source and binary forms, with or without
// modification, are permitted provided that the following conditions
// are met:
//  * Redistributions of source code must retain the above copyright
//    notice, this list of conditions and the following disclaimer.
//  * Redistributions in binary form must reproduce the above copyright
//    notice, this list of conditions and the following disclaimer in the
//    documentation and/or other materials provided with the distribution.
//  * Neither the name of NVIDIA CORPORATION nor the names of its
//    contributors may be used to endorse or promote products derived
//    from this software without specific prior written permission.
//
// THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS ``AS IS'' AND ANY
// EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
// IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR
// PURPOSE ARE DISCLAIMED.  IN NO EVENT SHALL THE COPYRIGHT OWNER OR
// CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL,
// EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO,
// PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR
// PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY
// OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
// (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
// OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
//
//----------------------------------------------------------------------------------
/*============================================================================
                    NVIDIA FXAA 3.11 by TIMOTHY LOTTES
------------------------------------------------------------------------------
            Modified for Urho3D then modified for upbge
============================================================================*/

/*==========================================================================*/
//
//                      Urho3D specific preparations
//
/*--------------------------------------------------------------------------*/

/*==========================================================================*/
//
//                      Shader Requirements
//
/*--------------------------------------------------------------------------*/

uniform sampler2D bgl_RenderedTexture;
uniform float bgl_RenderedTextureWidth;
uniform float bgl_RenderedTextureHeight;

in vec4 bgl_TexCoord;
out vec4 fragColor;

vec2 fxaaQualityRcpFrame = 1.0 / vec2(bgl_RenderedTextureWidth, bgl_RenderedTextureHeight);

#define FXAA_FAST_PIXEL_OFFSET 0

/*--------------------------------------------------------------------------*/

#if (FXAA_FAST_PIXEL_OFFSET == 1)
    #define Fxaa_vec2 ivec2
#else
    #define Fxaa_vec2 vec2
#endif

/*============================================================================
                        FXAA QUALITY - TUNING KNOBS
------------------------------------------------------------------------------
NOTE the other tuning knobs are now in the shader function inputs!
============================================================================*/
#ifndef FXAA_QUALITY_PRESET
    //
    // Choose the quality preset.
    // This needs to be compiled into the shader as it effects code.
    // Best option to include multiple presets is to 
    // in each shader define the preset, then include this file.
    //
    // OPTIONS
    // -----------------------------------------------------------------------
    // 10 to 15 - default medium dither (10=fastest, 15=highest quality)
    // 20 to 29 - less dither, more expensive (20=fastest, 29=highest quality)
    // 39       - no dither, very expensive 
    //
    // NOTES
    // -----------------------------------------------------------------------
    // 12 = slightly faster then FXAA 3.9 and higher edge quality (default)
    // 13 = about same speed as FXAA 3.9 and better than 12
    // 23 = closest to FXAA 3.9 visually and performance wise
    //  _ = the lowest digit is directly related to performance
    // _  = the highest digit is directly related to style
    // 
    #define FXAA_QUALITY_PRESET 39 // ULTRA QUALITY WARNING
#endif


/*============================================================================
                           FXAA QUALITY - PRESETS
============================================================================*/

/*============================================================================
                     FXAA QUALITY - MEDIUM DITHER PRESETS
============================================================================*/
#if (FXAA_QUALITY_PRESET == 10)
    #define FXAA_QUALITY_PS 3
    #define FXAA_QUALITY_P0 1.5
    #define FXAA_QUALITY_P1 3.0
    #define FXAA_QUALITY_P2 12.0
#endif
/*--------------------------------------------------------------------------*/
#if (FXAA_QUALITY_PRESET == 11)
    #define FXAA_QUALITY_PS 4
    #define FXAA_QUALITY_P0 1.0
    #define FXAA_QUALITY_P1 1.5
    #define FXAA_QUALITY_P2 3.0
    #define FXAA_QUALITY_P3 12.0
#endif
/*--------------------------------------------------------------------------*/
#if (FXAA_QUALITY_PRESET == 12)
    #define FXAA_QUALITY_PS 5
    #define FXAA_QUALITY_P0 1.0
    #define FXAA_QUALITY_P1 1.5
    #define FXAA_QUALITY_P2 2.0
    #define FXAA_QUALITY_P3 4.0
    #define FXAA_QUALITY_P4 12.0
#endif
/*--------------------------------------------------------------------------*/
#if (FXAA_QUALITY_PRESET == 13)
    #define FXAA_QUALITY_PS 6
    #define FXAA_QUALITY_P0 1.0
    #define FXAA_QUALITY_P1 1.5
    #define FXAA_QUALITY_P2 2.0
    #define FXAA_QUALITY_P3 2.0
    #define FXAA_QUALITY_P4 4.0
    #define FXAA_QUALITY_P5 12.0
#endif
/*--------------------------------------------------------------------------*/
#if (FXAA_QUALITY_PRESET == 14)
    #define FXAA_QUALITY_PS 7
    #define FXAA_QUALITY_P0 1.0
    #define FXAA_QUALITY_P1 1.5
    #define FXAA_QUALITY_P2 2.0
    #define FXAA_QUALITY_P3 2.0
    #define FXAA_QUALITY_P4 2.0
    #define FXAA_QUALITY_P5 4.0
    #define FXAA_QUALITY_P6 12.0
#endif
/*--------------------------------------------------------------------------*/
#if (FXAA_QUALITY_PRESET == 15)
    #define FXAA_QUALITY_PS 8
    #define FXAA_QUALITY_P0 1.0
    #define FXAA_QUALITY_P1 1.5
    #define FXAA_QUALITY_P2 2.0
    #define FXAA_QUALITY_P3 2.0
    #define FXAA_QUALITY_P4 2.0
    #define FXAA_QUALITY_P5 2.0
    #define FXAA_QUALITY_P6 4.0
    #define FXAA_QUALITY_P7 12.0
#endif

/*============================================================================
                     FXAA QUALITY - LOW DITHER PRESETS
============================================================================*/
#if (FXAA_QUALITY_PRESET == 20)
    #define FXAA_QUALITY_PS 3
    #define FXAA_QUALITY_P0 1.5
    #define FXAA_QUALITY_P1 2.0
    #define FXAA_QUALITY_P2 8.0
#endif
/*--------------------------------------------------------------------------*/
#if (FXAA_QUALITY_PRESET == 21)
    #define FXAA_QUALITY_PS 4
    #define FXAA_QUALITY_P0 1.0
    #define FXAA_QUALITY_P1 1.5
    #define FXAA_QUALITY_P2 2.0
    #define FXAA_QUALITY_P3 8.0
#endif
/*--------------------------------------------------------------------------*/
#if (FXAA_QUALITY_PRESET == 22)
    #define FXAA_QUALITY_PS 5
    #define FXAA_QUALITY_P0 1.0
    #define FXAA_QUALITY_P1 1.5
    #define FXAA_QUALITY_P2 2.0
    #define FXAA_QUALITY_P3 2.0
    #define FXAA_QUALITY_P4 8.0
#endif
/*--------------------------------------------------------------------------*/
#if (FXAA_QUALITY_PRESET == 23)
    #define FXAA_QUALITY_PS 6
    #define FXAA_QUALITY_P0 1.0
    #define FXAA_QUALITY_P1 1.5
    #define FXAA_QUALITY_P2 2.0
    #define FXAA_QUALITY_P3 2.0
    #define FXAA_QUALITY_P4 2.0
    #define FXAA_QUALITY_P5 8.0
#endif
/*--------------------------------------------------------------------------*/
#if (FXAA_QUALITY_PRESET == 24)
    #define FXAA_QUALITY_PS 7
    #define FXAA_QUALITY_P0 1.0
    #define FXAA_QUALITY_P1 1.5
    #define FXAA_QUALITY_P2 2.0
    #define FXAA_QUALITY_P3 2.0
    #define FXAA_QUALITY_P4 2.0
    #define FXAA_QUALITY_P5 3.0
    #define FXAA_QUALITY_P6 8.0
#endif
/*--------------------------------------------------------------------------*/
#if (FXAA_QUALITY_PRESET == 25)
    #define FXAA_QUALITY_PS 8
    #define FXAA_QUALITY_P0 1.0
    #define FXAA_QUALITY_P1 1.5
    #define FXAA_QUALITY_P2 2.0
    #define FXAA_QUALITY_P3 2.0
    #define FXAA_QUALITY_P4 2.0
    #define FXAA_QUALITY_P5 2.0
    #define FXAA_QUALITY_P6 4.0
    #define FXAA_QUALITY_P7 8.0
#endif
/*--------------------------------------------------------------------------*/
#if (FXAA_QUALITY_PRESET == 26)
    #define FXAA_QUALITY_PS 9
    #define FXAA_QUALITY_P0 1.0
    #define FXAA_QUALITY_P1 1.5
    #define FXAA_QUALITY_P2 2.0
    #define FXAA_QUALITY_P3 2.0
    #define FXAA_QUALITY_P4 2.0
    #define FXAA_QUALITY_P5 2.0
    #define FXAA_QUALITY_P6 2.0
    #define FXAA_QUALITY_P7 4.0
    #define FXAA_QUALITY_P8 8.0
#endif
/*--------------------------------------------------------------------------*/
#if (FXAA_QUALITY_PRESET == 27)
    #define FXAA_QUALITY_PS 10
    #define FXAA_QUALITY_P0 1.0
    #define FXAA_QUALITY_P1 1.5
    #define FXAA_QUALITY_P2 2.0
    #define FXAA_QUALITY_P3 2.0
    #define FXAA_QUALITY_P4 2.0
    #define FXAA_QUALITY_P5 2.0
    #define FXAA_QUALITY_P6 2.0
    #define FXAA_QUALITY_P7 2.0
    #define FXAA_QUALITY_P8 4.0
    #define FXAA_QUALITY_P9 8.0
#endif
/*--------------------------------------------------------------------------*/
#if (FXAA_QUALITY_PRESET == 28)
    #define FXAA_QUALITY_PS 11
    #define FXAA_QUALITY_P0 1.0
    #define FXAA_QUALITY_P1 1.5
    #define FXAA_QUALITY_P2 2.0
    #define FXAA_QUALITY_P3 2.0
    #define FXAA_QUALITY_P4 2.0
    #define FXAA_QUALITY_P5 2.0
    #define FXAA_QUALITY_P6 2.0
    #define FXAA_QUALITY_P7 2.0
    #define FXAA_QUALITY_P8 2.0
    #define FXAA_QUALITY_P9 4.0
    #define FXAA_QUALITY_P10 8.0
#endif
/*--------------------------------------------------------------------------*/
#if (FXAA_QUALITY_PRESET == 29)
    #define FXAA_QUALITY_PS 12
    #define FXAA_QUALITY_P0 1.0
    #define FXAA_QUALITY_P1 1.5
    #define FXAA_QUALITY_P2 2.0
    #define FXAA_QUALITY_P3 2.0
    #define FXAA_QUALITY_P4 2.0
    #define FXAA_QUALITY_P5 2.0
    #define FXAA_QUALITY_P6 2.0
    #define FXAA_QUALITY_P7 2.0
    #define FXAA_QUALITY_P8 2.0
    #define FXAA_QUALITY_P9 2.0
    #define FXAA_QUALITY_P10 4.0
    #define FXAA_QUALITY_P11 8.0
#endif

/*============================================================================
                     FXAA QUALITY - EXTREME QUALITY
============================================================================*/
#if (FXAA_QUALITY_PRESET == 39)
    #define FXAA_QUALITY_PS 12
    #define FXAA_QUALITY_P0 1.0
    #define FXAA_QUALITY_P1 1.0
    #define FXAA_QUALITY_P2 1.0
    #define FXAA_QUALITY_P3 1.0
    #define FXAA_QUALITY_P4 1.0
    #define FXAA_QUALITY_P5 1.5
    #define FXAA_QUALITY_P6 2.0
    #define FXAA_QUALITY_P7 2.0
    #define FXAA_QUALITY_P8 2.0
    #define FXAA_QUALITY_P9 2.0
    #define FXAA_QUALITY_P10 4.0
    #define FXAA_QUALITY_P11 8.0
#endif

/*============================================================================
                            Support Functions
============================================================================*/

float CalcLuma(vec3 rgb)
{
    vec3 luma = vec3(0.299, 0.587, 0.114);
    return dot(rgb, luma);
}

/*--------------------------------------------------------------------------*/

#ifdef GL3

#define FxaaTexTop(t, p) vec4(textureLod(t, p, 0.0).rgb, 1.0)
#define LumaTop(t, p) CalcLuma(textureLod(t, p, 0.0).rgb)
#if (FXAA_FAST_PIXEL_OFFSET == 1)
    #define LumaOff(t, p, o, r) CalcLuma(textureLodOffset(t, p, 0.0, o).rgb)
#else
    #define LumaOff(t, p, o, r) CalcLuma(texture(t, p + (o * r), 0.0).rgb)
#endif

#else

#define FxaaTexTop(t, p) vec4(texture(t, p, 0.0).rgb, 1.0)
#define LumaTop(t, p) CalcLuma(texture(t, p, 0.0).rgb)
#if (FXAA_FAST_PIXEL_OFFSET == 1)
    #define LumaOff(t, p, o, r) CalcLuma(texture2DLodOffset(t, p, 0.0, o).rgb)
#else
    #define LumaOff(t, p, o, r) CalcLuma(texture(t, p + (o * r), 0.0).rgb)
#endif

#endif

/*============================================================================
                             FXAA3 QUALITY - PC
============================================================================*/
vec4 FxaaPixelShader(
    //
    // Use noperspective interpolation here (turn off perspective interpolation).
    // {xy} = center of pixel
    vec2 pos,
    //
    // Input color texture.
    // {rgb_} = color in linear or perceptual color space
    // if (FXAA_GREEN_AS_LUMA == 0)
    //     {__a} = luma in perceptual color space (not linear)
    sampler2D tex,
    //
    // Only used on FXAA Quality.
    // This must be from a constant/uniform.
    // {x_} = 1.0/screenWidthInPixels
    // {_y} = 1.0/screenHeightInPixels
    vec2 fxaaQualityRcpFrame,
    //
    // Only used on FXAA Quality.
    // This used to be the FXAA_QUALITY_SUBPIX define.
    // It is here now to allow easier tuning.
    // Choose the amount of sub-pixel aliasing removal.
    // This can effect sharpness.
    //   1.00 - upper limit (softer)
    //   0.75 - default amount of filtering
    //   0.50 - lower limit (sharper, less sub-pixel aliasing removal)
    //   0.25 - almost off
    //   0.00 - completely off
    float fxaaQualitySubpix,
    //
    // Only used on FXAA Quality.
    // This used to be the FXAA_QUALITY_EDGE_THRESHOLD define.
    // It is here now to allow easier tuning.
    // The minimum amount of local contrast required to apply algorithm.
    //   0.333 - too little (faster)
    //   0.250 - low quality
    //   0.166 - default
    //   0.125 - high quality 
    //   0.063 - overkill (slower)
    float fxaaQualityEdgeThreshold,
    //
    // Only used on FXAA Quality.
    // This used to be the FXAA_QUALITY_EDGE_THRESHOLD_MIN define.
    // It is here now to allow easier tuning.
    // Trims the algorithm from processing darks.
    //   0.0833 - upper limit (default, the start of visible unfiltered edges)
    //   0.0625 - high quality (faster)
    //   0.0312 - visible limit (slower)
    // Special notes when using FXAA_GREEN_AS_LUMA,
    //   Likely want to set this to zero.
    //   As colors that are mostly not-green
    //   will appear very dark in the green channel!
    //   Tune by looking at mostly non-green content,
    //   then start at zero and increase until aliasing is a problem.
    float fxaaQualityEdgeThresholdMin
) {
/*--------------------------------------------------------------------------*/
    vec2 posM;
    posM.x = pos.x;
    posM.y = pos.y;
    
    vec4 rgbyM = FxaaTexTop(tex, posM);
    rgbyM.y = CalcLuma(rgbyM.rgb);
    #define lumaM rgbyM.y
    float lumaS = LumaOff(tex, posM, Fxaa_vec2( 0, 1), fxaaQualityRcpFrame.xy);
    float lumaE = LumaOff(tex, posM, Fxaa_vec2( 1, 0), fxaaQualityRcpFrame.xy);
    float lumaN = LumaOff(tex, posM, Fxaa_vec2( 0,-1), fxaaQualityRcpFrame.xy);
    float lumaW = LumaOff(tex, posM, Fxaa_vec2(-1, 0), fxaaQualityRcpFrame.xy);
/*--------------------------------------------------------------------------*/
    float maxSM = max(lumaS, lumaM);
    float minSM = min(lumaS, lumaM);
    float maxESM = max(lumaE, maxSM);
    float minESM = min(lumaE, minSM);
    float maxWN = max(lumaN, lumaW);
    float minWN = min(lumaN, lumaW);
    float rangeMax = max(maxWN, maxESM);
    float rangeMin = min(minWN, minESM);
    float rangeMaxScaled = rangeMax * fxaaQualityEdgeThreshold;
    float range = rangeMax - rangeMin;
    float rangeMaxClamped = max(fxaaQualityEdgeThresholdMin, rangeMaxScaled);
    bool earlyExit = range < rangeMaxClamped;
/*--------------------------------------------------------------------------*/
    if(earlyExit)
        return FxaaTexTop(tex, pos);
/*--------------------------------------------------------------------------*/
    float lumaNW = LumaOff(tex, posM, Fxaa_vec2(-1,-1), fxaaQualityRcpFrame.xy);
    float lumaSE = LumaOff(tex, posM, Fxaa_vec2( 1, 1), fxaaQualityRcpFrame.xy);
    float lumaNE = LumaOff(tex, posM, Fxaa_vec2( 1,-1), fxaaQualityRcpFrame.xy);
    float lumaSW = LumaOff(tex, posM, Fxaa_vec2(-1, 1), fxaaQualityRcpFrame.xy);
/*--------------------------------------------------------------------------*/
    float lumaNS = lumaN + lumaS;
    float lumaWE = lumaW + lumaE;
    float subpixRcpRange = 1.0/range;
    float subpixNSWE = lumaNS + lumaWE;
    float edgeHorz1 = (-2.0 * lumaM) + lumaNS;
    float edgeVert1 = (-2.0 * lumaM) + lumaWE;
/*--------------------------------------------------------------------------*/
    float lumaNESE = lumaNE + lumaSE;
    float lumaNWNE = lumaNW + lumaNE;
    float edgeHorz2 = (-2.0 * lumaE) + lumaNESE;
    float edgeVert2 = (-2.0 * lumaN) + lumaNWNE;
/*--------------------------------------------------------------------------*/
    float lumaNWSW = lumaNW + lumaSW;
    float lumaSWSE = lumaSW + lumaSE;
    float edgeHorz4 = (abs(edgeHorz1) * 2.0) + abs(edgeHorz2);
    float edgeVert4 = (abs(edgeVert1) * 2.0) + abs(edgeVert2);
    float edgeHorz3 = (-2.0 * lumaW) + lumaNWSW;
    float edgeVert3 = (-2.0 * lumaS) + lumaSWSE;
    float edgeHorz = abs(edgeHorz3) + edgeHorz4;
    float edgeVert = abs(edgeVert3) + edgeVert4;
/*--------------------------------------------------------------------------*/
    float subpixNWSWNESE = lumaNWSW + lumaNESE;
    float lengthSign = fxaaQualityRcpFrame.x;
    bool horzSpan = edgeHorz >= edgeVert;
    float subpixA = subpixNSWE * 2.0 + subpixNWSWNESE;
/*--------------------------------------------------------------------------*/
    if(!horzSpan) lumaN = lumaW;
    if(!horzSpan) lumaS = lumaE;
    if(horzSpan) lengthSign = fxaaQualityRcpFrame.y;
    float subpixB = (subpixA * (1.0/12.0)) - lumaM;
/*--------------------------------------------------------------------------*/
    float gradientN = lumaN - lumaM;
    float gradientS = lumaS - lumaM;
    float lumaNN = lumaN + lumaM;
    float lumaSS = lumaS + lumaM;
    bool pairN = abs(gradientN) >= abs(gradientS);
    float gradient = max(abs(gradientN), abs(gradientS));
    if(pairN) lengthSign = -lengthSign;
    float subpixC = clamp((abs(subpixB) * subpixRcpRange), 0.0, 1.0);
/*--------------------------------------------------------------------------*/
    vec2 posB;
    posB.x = posM.x;
    posB.y = posM.y;
    vec2 offNP;
    offNP.x = (!horzSpan) ? 0.0 : fxaaQualityRcpFrame.x;
    offNP.y = ( horzSpan) ? 0.0 : fxaaQualityRcpFrame.y;
    if(!horzSpan) posB.x += lengthSign * 0.5;
    if( horzSpan) posB.y += lengthSign * 0.5;
/*--------------------------------------------------------------------------*/
    vec2 posN;
    posN.x = posB.x - offNP.x * FXAA_QUALITY_P0;
    posN.y = posB.y - offNP.y * FXAA_QUALITY_P0;
    vec2 posP;
    posP.x = posB.x + offNP.x * FXAA_QUALITY_P0;
    posP.y = posB.y + offNP.y * FXAA_QUALITY_P0;
    float subpixD = ((-2.0)*subpixC) + 3.0;
    float lumaEndN = LumaTop(tex, posN);
    float subpixE = subpixC * subpixC;
    float lumaEndP = LumaTop(tex, posP);
/*--------------------------------------------------------------------------*/
    if(!pairN) lumaNN = lumaSS;
    float gradientScaled = gradient * 1.0/4.0;
    float lumaMM = lumaM - lumaNN * 0.5;
    float subpixF = subpixD * subpixE;
    bool lumaMLTZero = lumaMM < 0.0;
/*--------------------------------------------------------------------------*/
    lumaEndN -= lumaNN * 0.5;
    lumaEndP -= lumaNN * 0.5;
    bool doneN = abs(lumaEndN) >= gradientScaled;
    bool doneP = abs(lumaEndP) >= gradientScaled;
    if(!doneN) posN.x -= offNP.x * FXAA_QUALITY_P1;
    if(!doneN) posN.y -= offNP.y * FXAA_QUALITY_P1;
    bool doneNP = (!doneN) || (!doneP);
    if(!doneP) posP.x += offNP.x * FXAA_QUALITY_P1;
    if(!doneP) posP.y += offNP.y * FXAA_QUALITY_P1;
/*--------------------------------------------------------------------------*/
    if(doneNP) {
        if(!doneN) lumaEndN = LumaTop(tex, posN.xy);
        if(!doneP) lumaEndP = LumaTop(tex, posP.xy);
        if(!doneN) lumaEndN = lumaEndN - lumaNN * 0.5;
        if(!doneP) lumaEndP = lumaEndP - lumaNN * 0.5;
        doneN = abs(lumaEndN) >= gradientScaled;
        doneP = abs(lumaEndP) >= gradientScaled;
        if(!doneN) posN.x -= offNP.x * FXAA_QUALITY_P2;
        if(!doneN) posN.y -= offNP.y * FXAA_QUALITY_P2;
        doneNP = (!doneN) || (!doneP);
        if(!doneP) posP.x += offNP.x * FXAA_QUALITY_P2;
        if(!doneP) posP.y += offNP.y * FXAA_QUALITY_P2;
/*--------------------------------------------------------------------------*/
        #if (FXAA_QUALITY_PS > 3)
        if(doneNP) {
            if(!doneN) lumaEndN = LumaTop(tex, posN.xy);
            if(!doneP) lumaEndP = LumaTop(tex, posP.xy);
            if(!doneN) lumaEndN = lumaEndN - lumaNN * 0.5;
            if(!doneP) lumaEndP = lumaEndP - lumaNN * 0.5;
            doneN = abs(lumaEndN) >= gradientScaled;
            doneP = abs(lumaEndP) >= gradientScaled;
            if(!doneN) posN.x -= offNP.x * FXAA_QUALITY_P3;
            if(!doneN) posN.y -= offNP.y * FXAA_QUALITY_P3;
            doneNP = (!doneN) || (!doneP);
            if(!doneP) posP.x += offNP.x * FXAA_QUALITY_P3;
            if(!doneP) posP.y += offNP.y * FXAA_QUALITY_P3;
/*--------------------------------------------------------------------------*/
            #if (FXAA_QUALITY_PS > 4)
            if(doneNP) {
                if(!doneN) lumaEndN = LumaTop(tex, posN.xy);
                if(!doneP) lumaEndP = LumaTop(tex, posP.xy);
                if(!doneN) lumaEndN = lumaEndN - lumaNN * 0.5;
                if(!doneP) lumaEndP = lumaEndP - lumaNN * 0.5;
                doneN = abs(lumaEndN) >= gradientScaled;
                doneP = abs(lumaEndP) >= gradientScaled;
                if(!doneN) posN.x -= offNP.x * FXAA_QUALITY_P4;
                if(!doneN) posN.y -= offNP.y * FXAA_QUALITY_P4;
                doneNP = (!doneN) || (!doneP);
                if(!doneP) posP.x += offNP.x * FXAA_QUALITY_P4;
                if(!doneP) posP.y += offNP.y * FXAA_QUALITY_P4;
/*--------------------------------------------------------------------------*/
                #if (FXAA_QUALITY_PS > 5)
                if(doneNP) {
                    if(!doneN) lumaEndN = LumaTop(tex, posN.xy);
                    if(!doneP) lumaEndP = LumaTop(tex, posP.xy);
                    if(!doneN) lumaEndN = lumaEndN - lumaNN * 0.5;
                    if(!doneP) lumaEndP = lumaEndP - lumaNN * 0.5;
                    doneN = abs(lumaEndN) >= gradientScaled;
                    doneP = abs(lumaEndP) >= gradientScaled;
                    if(!doneN) posN.x -= offNP.x * FXAA_QUALITY_P5;
                    if(!doneN) posN.y -= offNP.y * FXAA_QUALITY_P5;
                    doneNP = (!doneN) || (!doneP);
                    if(!doneP) posP.x += offNP.x * FXAA_QUALITY_P5;
                    if(!doneP) posP.y += offNP.y * FXAA_QUALITY_P5;
/*--------------------------------------------------------------------------*/
                    #if (FXAA_QUALITY_PS > 6)
                    if(doneNP) {
                        if(!doneN) lumaEndN = LumaTop(tex, posN.xy);
                        if(!doneP) lumaEndP = LumaTop(tex, posP.xy);
                        if(!doneN) lumaEndN = lumaEndN - lumaNN * 0.5;
                        if(!doneP) lumaEndP = lumaEndP - lumaNN * 0.5;
                        doneN = abs(lumaEndN) >= gradientScaled;
                        doneP = abs(lumaEndP) >= gradientScaled;
                        if(!doneN) posN.x -= offNP.x * FXAA_QUALITY_P6;
                        if(!doneN) posN.y -= offNP.y * FXAA_QUALITY_P6;
                        doneNP = (!doneN) || (!doneP);
                        if(!doneP) posP.x += offNP.x * FXAA_QUALITY_P6;
                        if(!doneP) posP.y += offNP.y * FXAA_QUALITY_P6;
/*--------------------------------------------------------------------------*/
                        #if (FXAA_QUALITY_PS > 7)
                        if(doneNP) {
                            if(!doneN) lumaEndN = LumaTop(tex, posN.xy);
                            if(!doneP) lumaEndP = LumaTop(tex, posP.xy);
                            if(!doneN) lumaEndN = lumaEndN - lumaNN * 0.5;
                            if(!doneP) lumaEndP = lumaEndP - lumaNN * 0.5;
                            doneN = abs(lumaEndN) >= gradientScaled;
                            doneP = abs(lumaEndP) >= gradientScaled;
                            if(!doneN) posN.x -= offNP.x * FXAA_QUALITY_P7;
                            if(!doneN) posN.y -= offNP.y * FXAA_QUALITY_P7;
                            doneNP = (!doneN) || (!doneP);
                            if(!doneP) posP.x += offNP.x * FXAA_QUALITY_P7;
                            if(!doneP) posP.y += offNP.y * FXAA_QUALITY_P7;
/*--------------------------------------------------------------------------*/
    #if (FXAA_QUALITY_PS > 8)
    if(doneNP) {
        if(!doneN) lumaEndN = LumaTop(tex, posN.xy);
        if(!doneP) lumaEndP = LumaTop(tex, posP.xy);
        if(!doneN) lumaEndN = lumaEndN - lumaNN * 0.5;
        if(!doneP) lumaEndP = lumaEndP - lumaNN * 0.5;
        doneN = abs(lumaEndN) >= gradientScaled;
        doneP = abs(lumaEndP) >= gradientScaled;
        if(!doneN) posN.x -= offNP.x * FXAA_QUALITY_P8;
        if(!doneN) posN.y -= offNP.y * FXAA_QUALITY_P8;
        doneNP = (!doneN) || (!doneP);
        if(!doneP) posP.x += offNP.x * FXAA_QUALITY_P8;
        if(!doneP) posP.y += offNP.y * FXAA_QUALITY_P8;
/*--------------------------------------------------------------------------*/
        #if (FXAA_QUALITY_PS > 9)
        if(doneNP) {
            if(!doneN) lumaEndN = LumaTop(tex, posN.xy);
            if(!doneP) lumaEndP = LumaTop(tex, posP.xy);
            if(!doneN) lumaEndN = lumaEndN - lumaNN * 0.5;
            if(!doneP) lumaEndP = lumaEndP - lumaNN * 0.5;
            doneN = abs(lumaEndN) >= gradientScaled;
            doneP = abs(lumaEndP) >= gradientScaled;
            if(!doneN) posN.x -= offNP.x * FXAA_QUALITY_P9;
            if(!doneN) posN.y -= offNP.y * FXAA_QUALITY_P9;
            doneNP = (!doneN) || (!doneP);
            if(!doneP) posP.x += offNP.x * FXAA_QUALITY_P9;
            if(!doneP) posP.y += offNP.y * FXAA_QUALITY_P9;
/*--------------------------------------------------------------------------*/
            #if (FXAA_QUALITY_PS > 10)
            if(doneNP) {
                if(!doneN) lumaEndN = LumaTop(tex, posN.xy);
                if(!doneP) lumaEndP = LumaTop(tex, posP.xy);
                if(!doneN) lumaEndN = lumaEndN - lumaNN * 0.5;
                if(!doneP) lumaEndP = lumaEndP - lumaNN * 0.5;
                doneN = abs(lumaEndN) >= gradientScaled;
                doneP = abs(lumaEndP) >= gradientScaled;
                if(!doneN) posN.x -= offNP.x * FXAA_QUALITY_P10;
                if(!doneN) posN.y -= offNP.y * FXAA_QUALITY_P10;
                doneNP = (!doneN) || (!doneP);
                if(!doneP) posP.x += offNP.x * FXAA_QUALITY_P10;
                if(!doneP) posP.y += offNP.y * FXAA_QUALITY_P10;
/*--------------------------------------------------------------------------*/
                #if (FXAA_QUALITY_PS > 11)
                if(doneNP) {
                    if(!doneN) lumaEndN = LumaTop(tex, posN.xy);
                    if(!doneP) lumaEndP = LumaTop(tex, posP.xy);
                    if(!doneN) lumaEndN = lumaEndN - lumaNN * 0.5;
                    if(!doneP) lumaEndP = lumaEndP - lumaNN * 0.5;
                    doneN = abs(lumaEndN) >= gradientScaled;
                    doneP = abs(lumaEndP) >= gradientScaled;
                    if(!doneN) posN.x -= offNP.x * FXAA_QUALITY_P11;
                    if(!doneN) posN.y -= offNP.y * FXAA_QUALITY_P11;
                    doneNP = (!doneN) || (!doneP);
                    if(!doneP) posP.x += offNP.x * FXAA_QUALITY_P11;
                    if(!doneP) posP.y += offNP.y * FXAA_QUALITY_P11;
/*--------------------------------------------------------------------------*/
                    #if (FXAA_QUALITY_PS > 12)
                    if(doneNP) {
                        if(!doneN) lumaEndN = LumaTop(tex, posN.xy);
                        if(!doneP) lumaEndP = LumaTop(tex, posP.xy);
                        if(!doneN) lumaEndN = lumaEndN - lumaNN * 0.5;
                        if(!doneP) lumaEndP = lumaEndP - lumaNN * 0.5;
                        doneN = abs(lumaEndN) >= gradientScaled;
                        doneP = abs(lumaEndP) >= gradientScaled;
                        if(!doneN) posN.x -= offNP.x * FXAA_QUALITY_P12;
                        if(!doneN) posN.y -= offNP.y * FXAA_QUALITY_P12;
                        doneNP = (!doneN) || (!doneP);
                        if(!doneP) posP.x += offNP.x * FXAA_QUALITY_P12;
                        if(!doneP) posP.y += offNP.y * FXAA_QUALITY_P12;
/*--------------------------------------------------------------------------*/
                    }
                    #endif
/*--------------------------------------------------------------------------*/
                }
                #endif
/*--------------------------------------------------------------------------*/
            }
            #endif
/*--------------------------------------------------------------------------*/
        }
        #endif
/*--------------------------------------------------------------------------*/
    }
    #endif
/*--------------------------------------------------------------------------*/
                        }
                        #endif
/*--------------------------------------------------------------------------*/
                    }
                    #endif
/*--------------------------------------------------------------------------*/
                }
                #endif
/*--------------------------------------------------------------------------*/
            }
            #endif
/*--------------------------------------------------------------------------*/
        }
        #endif
/*--------------------------------------------------------------------------*/
    }
/*--------------------------------------------------------------------------*/
    float dstN = posM.x - posN.x;
    float dstP = posP.x - posM.x;
    if(!horzSpan) dstN = posM.y - posN.y;
    if(!horzSpan) dstP = posP.y - posM.y;
/*--------------------------------------------------------------------------*/
    bool goodSpanN = (lumaEndN < 0.0) != lumaMLTZero;
    float spanLength = (dstP + dstN);
    bool goodSpanP = (lumaEndP < 0.0) != lumaMLTZero;
    float spanLengthRcp = 1.0/spanLength;
/*--------------------------------------------------------------------------*/
    bool directionN = dstN < dstP;
    float dst = min(dstN, dstP);
    bool goodSpan = directionN ? goodSpanN : goodSpanP;
    float subpixG = subpixF * subpixF;
    float pixelOffset = (dst * (-spanLengthRcp)) + 0.5;
    float subpixH = subpixG * fxaaQualitySubpix;
/*--------------------------------------------------------------------------*/
    float pixelOffsetGood = goodSpan ? pixelOffset : 0.0;
    float pixelOffsetSubpix = max(pixelOffsetGood, subpixH);
    if(!horzSpan) posM.x += pixelOffsetSubpix * lengthSign;
    if( horzSpan) posM.y += pixelOffsetSubpix * lengthSign;
    return FxaaTexTop(tex, posM);
}
/*==========================================================================*/

/*============================================================================
                      Urho3D Vertex- and Pixelshader
                      
============================================================================*/


void main()
{

    vec2 cGBufferInvSize = fxaaQualityRcpFrame;
    vec2 rcpFrame = vec2(cGBufferInvSize.x, cGBufferInvSize.y);
    
    vec2 vScreenPos = bgl_TexCoord.xy;

    fragColor = FxaaPixelShader(
        vScreenPos,                         // vec2 pos,
        bgl_RenderedTexture,                // sampler2D tex,
        rcpFrame,                           // vec2 fxaaQualityRcpFrame,
        0.125,                               // float fxaaQualitySubpix,
        0.166,                              // float fxaaQualityEdgeThreshold,
        0.0833                              // float fxaaQualityEdgeThresholdMin
    );
}
"""


class FXAA(ULFilter):

    def __init__(self, idx: int = None) -> None:
        super().__init__(glsl, idx)



from bge import logic
from mathutils import Vector, Matrix
from uplogic.utils.errors import PassIndexOccupiedError
from uplogic.utils import debug
from pathlib import Path


def load_glsl(filepath: str):
    return Path(filepath).read_text()


class ULFilter():
    '''Wrapper for KX_2DFilter.

    :param `program`: GLSL code as `str`.
    :param `idx`: Pass Index for this filter.
    :param `uniforms`: A `dict` of [`str`: `dict`] binding dictionary values to
    the filter in the form "key of dictionary".
    '''

    @property
    def active(self):
        return self in FilterSystem.filters.values()
    
    @active.setter
    def active(self, val):
        if val:
            self.startup()
        else:
            self.shutdown()


    def __init__(
        self,
        program: str,
        idx: int = None,
        bound_uniforms: dict = {}
    ) -> None:
        self.program = program
        self.idx = idx
        scene = logic.getCurrentScene()
        self.manager = scene.filterManager
        self._uniforms = bound_uniforms
        self._filter = None
        FilterSystem.add_filter(self)

    def startup(self):
        if FilterSystem.filters.get(self.idx) is not None:
            raise PassIndexOccupiedError(self.idx)
        FilterSystem.filters[self.idx] = self
        self._filter = self.manager.addFilter(self.idx, 12, self.program)
        uniforms = self._uniforms
        for uniform in uniforms:
            self.set_uniform(uniform, uniforms[uniform].get(uniform))
        if uniforms.keys():
            logic.getCurrentScene().post_draw.append(self.update)

    def update(self):
        uniforms = self._uniforms
        for uniform in uniforms:
            self.set_uniform(uniform, uniforms[uniform].get(uniform))

    def shutdown(self):
        if self.update in logic.getCurrentScene().post_draw:
            logic.getCurrentScene().post_draw.remove(self.update)
        if self in FilterSystem.filters.values():
            FilterSystem.filters.pop(self.idx)
            self.manager.removeFilter(self.idx)

    def set_uniform(self, name, value):
        cls = value.__class__
        if cls is int:
            self._filter.setUniform1i(name, value)
        elif cls is float:
            self._filter.setUniform1f(name, value)
        elif cls is Vector:
            dim = len(value)
            if dim == 2:
                self._filter.setUniform2f(name, value.x, value.y)
            if dim == 3:
                self._filter.setUniform3f(name, value.x, value.y, value.z)
            if dim == 4:
                self._filter.setUniform4f(name, value.x, value.y, value.z, value.w)
        elif cls is Matrix:
            rows = len(value.row)
            cols = len(value.col)
            if rows == cols == 3:
                self._filter.setUniformMatrix3(
                    name,(
                        [value[0][0], value[0][1], value[0][2]],
                        [value[1][0], value[1][1], value[1][2]],
                        [value[2][0], value[2][1], value[2][2]]
                    ),
                    False
                )
            elif rows == cols == 4:
                self._filter.setUniformMatrix4(
                    name,
                    (
                        [value[0][0], value[0][1], value[0][2], value[0][3]],
                        [value[1][0], value[1][1], value[1][2], value[1][3]],
                        [value[2][0], value[2][1], value[2][2], value[2][3]],
                        [value[3][0], value[2][1], value[3][2], value[3][3]]
                    ),
                )


class FilterSystem:
    filters: dict[ULFilter] = {}

    @classmethod
    def get_filter(cls, idx):
        return logic.getCurrentScene().filterManager.getFilter(idx)

    @classmethod
    def add_filter(cls, filter):
        if filter.idx and cls.filters.get(filter.idx, None) is None:
            filter.startup()
        elif filter.idx is not None and cls.filters.get(filter.idx):
            #raise PassIndexOccupiedError
            debug(f"2D Filter pass index {filter.idx} already in-use!")
        else:
            idx = 0
            while cls.filters.get(idx, None) is not None:
                idx += 1
            filter.idx = idx
            filter.startup()

    @classmethod
    def remove_filter(cls, filter):
        if isinstance(filter, int) and cls.filters.get(filter, None):
            cls.filters.get(filter).shutdown()


from bge import logic
from uplogic.nodes import ULParameterNode
from uplogic.nodes import ULOutSocket
from uplogic.utils import is_invalid


class ULGamepadSticks(ULParameterNode):
    def __init__(self, axis=0):
        ULParameterNode.__init__(self)
        self.axis = axis
        self.inverted = None
        self.index = None
        self.sensitivity = None
        self.threshold = None
        self._x_axis_values = None
        self._y_axis_values = None
        self.raw_values = [0, 0]
        self.X = ULOutSocket(self, self.get_x_axis)
        self.Y = ULOutSocket(self, self.get_y_axis)

    def get_x_axis(self):
        x = self.raw_values[0]
        if -self.threshold < x < self.threshold:
            x = 0
        return x * self._sensitivity

    def get_y_axis(self):
        y = self.raw_values[1]
        if -self.threshold < y < self.threshold:
            y = 0
        return y * self._sensitivity

    def evaluate(self):
        self._set_ready()
        index = self.get_input(self.index)

        if logic.joysticks[index]:
            joystick = logic.joysticks[index]
        else:
            self._x_axis_values = 0
            self._y_axis_values = 0
            return
        if is_invalid(joystick):
            return
        axis = self.get_input(self.axis)
        raw_values = joystick.axisValues
        if axis == 0:
            self.raw_values = [raw_values[0], raw_values[1]]
        elif axis == 1:
            self.raw_values = [raw_values[2], raw_values[3]]
        inverted = self.get_input(self.inverted)
        sensitivity = self.get_input(self.sensitivity)
        self._sensitivity = -sensitivity if inverted else sensitivity
        self.threshold = self.get_input(self.threshold)



from bge import render

def draw_line(start, end, color=[1, 1, 1, 1]):
    render.drawLine(
        start,
        end,
        color
    )

def draw_cube(origin, width, color=[1, 1, 1, 1]):
    draw_box(origin, width, width, width, color)


def draw_box(origin, length, width, height, color=[1, 1, 1, 1]):
    c1 = origin.copy()
    c2 = origin.copy()
    c3 = origin.copy()
    c4 = origin.copy()
    c5 = origin.copy()
    c6 = origin.copy()
    c7 = origin.copy()

    c1[0] += width

    c2[0] += width
    c2[1] += length

    c3[1] += length

    c4[2] += height

    c5[0] += width
    c5[2] += height

    c6[0] += width
    c6[1] += length
    c6[2] += height

    c7[1] += length
    c7[2] += height

    draw_line(origin, c1, color)
    draw_line(c1, c2, color)
    draw_line(c2, c3, color)
    draw_line(c3, origin, color)

    draw_line(origin, c4, color)
    draw_line(c1, c5, color)
    draw_line(c2, c6, color)
    draw_line(c3, c7, color)

    draw_line(c4, c5, color)
    draw_line(c5, c6, color)
    draw_line(c6, c7, color)
    draw_line(c7, c4, color)



from .shader import ULFilter
from bge import logic


glsl = """
// Shader authored by LieGri (https://blenderartists.org/t/bge-hbao-ambient-occlusion-shader/690374)
// Modified for UPBGE 0.3+ by Iza Zed

uniform sampler2D bgl_DepthTexture;
uniform sampler2D bgl_RenderedTexture;
uniform float bgl_RenderedTextureWidth;
uniform float bgl_RenderedTextureHeight;
in vec4 bgl_TexCoord;
out vec4 fragColor;

//THIS NEEDS TO MATCH YOUR CAMERA SETTINGS---------------------
uniform float znear;                    //Z-near
uniform float zfar;                     //Z-far
uniform float fov;                      //FoV
//-------------------------------------------------------------

//USER VARIABLES-----------------------------------------------
const float intensity = 1.2;            //Intensity of the AO effect
const float sampleRadius = 4.0;         //Radius of the AO, bigger values need more performance


const float sampleDirections = 10.0;    //Main sample count, affects performance heavily 
const float sampleSteps = 32.0;         //SubSample count, sometimes higher values are faster


const bool useAttenuation = true;      //Applies attenuation to each AO sample, different values look better depending on your scene
const float attenuationScale = 0.7;     //Depth scale of the attenuation, different values look better depending on your scene


const float angleBias = 0.2;            //Brightens up the AO effect to reduce noise and artifacts


const bool noise = true;               //Use noise instead of pattern for sample dithering, much slower but more accurate
const float noiseamount = .1;          //Per-Pixel noise amount, bigger values need more performance
const float jitterAmount = .1;         //Per-Sample noise amount, bigger values need more performance


const float lumInfluence = 0.0;         //Influence of the luminance on the AO effect


const bool onlyAO = false;               //Only show AO pass for debugging
const bool externalBlur = false;         //Store AO in alpha pass for a later blur
//-------------------------------------------------------------


float width = bgl_RenderedTextureWidth;
float height = bgl_RenderedTextureHeight;


vec2 texCoord = bgl_TexCoord.xy;


float aspectratio = width/height;
float thfov = tan(fov * 0.0087266462597222);
const float TWO_PI = 6.283185307;


float linearize_depth(vec2 uv){
    float n = .1;
    float f = 100.0;
    float z = texture(bgl_DepthTexture, uv).x;
    return (2.0 * n) / (f + n - z * (f - n));
}


vec3 getLinearColor(vec2 coord)
{    
    vec3 C = vec3(texture(bgl_RenderedTexture, coord));
    C.r = pow(C.r, 2.2);
    C.g = pow(C.g, 2.2);
    C.b = pow(C.b, 2.2);
    return C.rgb;
}


vec3 sRGBToLinear(vec3 C)
{
    C.r = pow(C.r, 2.2);
    C.g = pow(C.g, 2.2);
    C.b = pow(C.b, 2.2);
    return C.rgb;
}


vec3 linearTosRGB(vec3 C)
{
    C.r = pow(C.r, 0.45454545);
    C.g = pow(C.g, 0.45454545);
    C.b = pow(C.b, 0.45454545);
    return C.rgb;
}


float getLinearDepth(vec2 coord)
{
    float zdepth = texture(bgl_DepthTexture,coord).x;
    return zfar*znear / (zfar + zdepth * (znear - zfar));
}


vec3 getViewVector(vec2 coord)
{
    vec2 ndc = (coord * 2.0 - 1.0);
    return vec3(ndc.x*thfov, ndc.y*thfov/aspectratio, 1.0);
}


vec3 getViewPosition(vec2 coord)
{
    return getViewVector(coord) * getLinearDepth(coord);
}


vec3 getViewNormal(vec2 coord)
{
    float pW = 1.0/width;
    float pH = 1.0/height;
    
    vec3 p1 = getViewPosition(coord+vec2(pW,0.0)).xyz;
    vec3 p2 = getViewPosition(coord+vec2(0.0,pH)).xyz;
    vec3 p3 = getViewPosition(coord+vec2(-pW,0.0)).xyz;
    vec3 p4 = getViewPosition(coord+vec2(0.0,-pH)).xyz;


    vec3 vP = getViewPosition(coord);

    vec3 dx = vP-p1;
    vec3 dy = p2-vP;
    vec3 dx2 = p3-vP;
    vec3 dy2 = vP-p4;

    if(length(dx2)<length(dx)&&coord.x-pW>=0||coord.x+pW>1) {
    dx = dx2;
    }
    if(length(dy2)<length(dy)&&coord.y-pH>=0||coord.y+pH>1) {
    dy = dy2;
    }

    return normalize(cross( dx , dy ));
}


float rand(vec2 co)
{
    if (noise) {
        return fract(sin(dot(co.xy,vec2(12.9898,78.233)*3.0)) * 43758.5453);
    } else {
        return ((fract(1.0-co.s*(width/2.0))*0.3)+(fract(co.t*(height/2.0))*0.6))*2.0-1.0;
    }
}


vec2 rand2D(vec2 coord) //generating noise/pattern texture for dithering
{
    float noiseX = ((fract(1.0-coord.s*(width/2.0))*0.25)+(fract(coord.t*(height/2.0))*0.75))*2.0-1.0;
    float noiseY = ((fract(1.0-coord.s*(width/2.0))*0.75)+(fract(coord.t*(height/2.0))*0.25))*2.0-1.0;
    
    if (noise)
    {
        noiseX = clamp(fract(sin(dot(coord ,vec2(12.9898,78.233))) * 43758.5453),0.0,1.0)*2.0-1.0;
        noiseY = clamp(fract(sin(dot(coord ,vec2(12.9898,78.233)*2.0)) * 43758.5453),0.0,1.0)*2.0-1.0;
    }
    return vec2(noiseX,noiseY)*noiseamount;
}


mat4 getViewProjectionMatrix()
{
    mat4 result;
    
    float frustumDepth = zfar - znear;
    float oneOverDepth = 1 / frustumDepth;


    result[0][0] = 1 / thfov;
    result[1][1] = aspectratio * result[0][0];
    result[2][2] = zfar * oneOverDepth;
    result[2][3] = 1;
    result[3][2] = (-zfar * znear) * oneOverDepth;

    return result;
}


vec2 ComputeFOVProjection(vec3 pos, mat4 VPM)
{
    vec4 offset = vec4(pos, 1.0);
    offset = VPM * offset;
    offset.xy /= offset.w;
    return offset.xy * 0.5 + 0.5;
}


void main()
{
    mat4 VPM = getViewProjectionMatrix();
    mat4 viewProjectionInverseMatrix  = inverse(VPM);

    vec3 color = texture(bgl_RenderedTexture, bgl_TexCoord.xy).rgb;
    float d = linearize_depth(texCoord);
    if (d >= 0.9){
        fragColor.rgb = color;
        fragColor.a = 1;
        return;
    }

    vec3 originVS = getViewPosition(texCoord);
    vec3 normalVS = getViewNormal(texCoord);

    float radiusSS = 0.0;
    float radiusWS = 0.0;

    radiusSS = sampleRadius;
    vec4 temp0 = viewProjectionInverseMatrix * vec4(0.0, 0.0, -1.0, 1.0);
    vec3 out0 = temp0.xyz;
    vec4 temp1 = viewProjectionInverseMatrix * vec4(radiusSS, 0.0, -1.0, 1.0);
    vec3 out1 = temp1.xyz;

    radiusWS = min(tan(radiusSS * fov * 0.0087266462597222) * originVS.z, length(out1 - out0));

    /*if (radiusSS < 1.0 / width) {
        fragColor.rgb = color;
        fragColor.a = 1.0;
        return;
    }*/

    const float theta = TWO_PI / float(sampleDirections);
    float cosTheta = cos(theta);
    float sinTheta = sin(theta);

    mat2 deltaRotationMatrix = mat2(cosTheta, -sinTheta, sinTheta, cosTheta);
    vec2 deltaUV = vec2(1.0, 0.0) * (radiusSS / (float(sampleDirections * sampleSteps) + 1.0));

    vec2 sampleNoise = rand2D(texCoord);
    mat2 rotationMatrix = mat2(sampleNoise.x, -sampleNoise.y, 
                               sampleNoise.y,  sampleNoise.x);

    deltaUV = rotationMatrix * deltaUV;

    float occlusion = 0.0;
    float jitter = rand(texCoord) * jitterAmount;

    for(int i = 0; i < sampleDirections; i++) {
        deltaUV = deltaRotationMatrix * deltaUV;

        vec2 sampleDirUV = deltaUV;
        float oldAngle = angleBias;

        for(int j = 0; j < sampleDirections; j++) {
            vec2 sampleUV = texCoord + (jitter + float(j)) * sampleDirUV;
            vec3 sampleVS = getViewPosition(sampleUV);
            vec3 sampleDirVS = (sampleVS - originVS);
            
            float gamma = 1.570796326 - acos(dot(normalVS, normalize(sampleDirVS)));
            
            if (gamma > oldAngle) {
                float value = sin(gamma) - sin(oldAngle);
                
                if(useAttenuation) {
                    float attenuation = clamp(1.0 - pow(length(sampleDirVS) / radiusWS * attenuationScale, 2.0), 0.0, 1.0);
                    occlusion += attenuation * value;
                } else {
                    occlusion += value;
                }
                
                oldAngle = gamma;
            }
        }
    }

    occlusion = 1.0 - (occlusion * intensity) / float(sampleDirections);
    occlusion = clamp(pow(occlusion, 1.0 + intensity), 0.0, 1.0);

    vec3 lumcoeff = vec3(0.299, 0.587, 0.114);
    float lum = dot(color.rgb, lumcoeff) * lumInfluence;

    occlusion = lum + ((1.0 - lum) * occlusion);
    occlusion = mix(occlusion, 1, d);

    fragColor.rgb = color * occlusion;
    fragColor.a = occlusion;

    if(externalBlur) {
        fragColor.rgb = color;
        fragColor.a = occlusion;
    }
    if(onlyAO) {
        fragColor.rgb = vec3(occlusion);
        fragColor.a = 1.0;
    }

    //fragColor.rgb = vec3(jitter);
}
"""


class HBAO(ULFilter):

    def __init__(self, idx: int = None) -> None:
        cam = logic.getCurrentScene().active_camera
        self.settings = {
            'znear': cam.near,
            'zfar': cam.far,
            'fov': cam.fov
        }
        super().__init__(glsl, idx, {
            'znear': self.settings,
            'zfar': self.settings,
            'fov': self.settings
        })

    def update(self):
        super().update()
        cam = logic.getCurrentScene().active_camera
        self.settings = {
            'znear': cam.near,
            'zfar': cam.far,
            'fov': cam.fov
        }




from bge import logic
from uplogic.animation.action import PLAY_MODES
import bpy
import time


class ULSequence():
    '''Play an image animation through a material node.
    
    :param `material`: Name of the material to play the animation on.
    :param `node`: Name of the node the image animation is loaded on.
    :param `start_frame`: Starting frame of the animation.
    :param `end_frame`: End frame of the animation.
    :param `fps`: Frames per second.
    :param `mode`: Animation mode, `str` of [`play`, `loop`, `pingpong`]
    '''

    def __init__(
        self,
        material: str,
        node: str,
        start_frame: int,
        end_frame: int,
        fps: int = 60,
        mode: str = 'play'
    ) -> None:
        self.material = material
        self.node = node
        self.start_frame = start_frame
        self.end_frame = end_frame
        self.fps = fps
        self.mode = PLAY_MODES.get(mode, 0)
        self.time = 0.0
        self.frame = 0
        self.initialized = False
        self.reverse = False
        self.running = True
        self._consumed = False
        self.active = True
        self._pause = False
        self._time_then = time.time()
        self.on_start = False
        self.on_finish = False
        self.player = (
            bpy.data.materials[material]
            .node_tree
            .nodes[node]
        ).image_user

        logic.getCurrentScene().pre_draw.append(self.update)

    def stop(self):
        '''Stop this animation completely.'''
        self.on_finish = True
        logic.getCurrentScene().pre_draw.remove(self.update)

    def pause(self):
        '''Pause this animation.'''
        self._pause = True
        self.running = False

    def restart(self):
        '''Restart this animation.'''
        self.initialized = False

    def unpause(self):
        '''Continue this animation.'''
        self._pause = False
        self.running = True

    def update(self):
        '''This is called each frame.'''
        now = time.time()
        player = self.player
        self.time += now - self._time_then
        self._time_then = now
        fps = self.fps
        rate = 1 / fps
        speed = self.time / rate
        if speed < 1:
            return
        self.time -= rate * speed
        if self._pause:
            return
        play_mode = self.mode
        running = self.running
        start_frame = self.end_frame if self.reverse else self.start_frame
        end_frame = self.start_frame if self.reverse else self.end_frame
        if not self.initialized:
            player.frame_offset = start_frame
            self.initialized = True
        inverted = (start_frame > end_frame)
        frame = self.frame = player.frame_offset
        reset_cond = (frame <= end_frame) if inverted else (frame >= end_frame)
        if not running:
            if reset_cond:
                player.frame_offset = start_frame if inverted else end_frame
            self.on_start = True
            self._consumed = False

        start_cond = frame > start_frame if inverted else frame < start_frame

        if start_cond:
            self.running = True
            player.frame_offset = start_frame
        frame = player.frame_offset
        run_cond = (frame > end_frame) if inverted else (frame < end_frame)
        if run_cond:
            self.running = True
            s = round(speed)
            if inverted:
                if frame - s < end_frame:
                    if play_mode == 1:
                        leftover = abs(frame - s - end_frame)
                        span = start_frame - end_frame
                        while leftover > span:
                            leftover -= span
                        player.frame_offset = start_frame - leftover
                    else:
                        player.frame_offset = end_frame
                else:
                    player.frame_offset -= s
            else:
                if frame + s > end_frame:
                    if play_mode == 1:
                        leftover = frame + s - end_frame
                        span = end_frame - start_frame
                        while leftover > span:
                            leftover -= span
                        player.frame_offset = start_frame + leftover
                    else:
                        player.frame_offset = end_frame
                else:
                    player.frame_offset += s
        elif play_mode == 1:
            player.frame_offset = end_frame if inverted else start_frame
        elif play_mode == 2:
            self.reverse = not self.reverse
        else:
            self.stop()



from mathutils import Euler
from mathutils import Vector
from uplogic.nodes import ULOutSocket
from uplogic.nodes import ULParameterNode


class ULMatrixToXYZ(ULParameterNode):
    def __init__(self):
        ULParameterNode.__init__(self)
        self.input_m = None
        self.output = None
        self.euler = Euler()
        self.OUT = ULOutSocket(self, self.get_out)

    def get_out(self):
        matrix = self.get_input(self.input_m)
        if self.output:
            return matrix.to_euler()
        else:
            e = matrix.to_euler()
            return Vector((e.x, e.y, e.z))

    def evaluate(self):
        self._set_ready()



from .shader import ULFilter


glsl = """
uniform sampler2D bgl_RenderedTexture;
in vec4 bgl_TexCoord;
uniform float brightness;

out vec4 fragColor;

void main() {
    fragColor = texture(bgl_RenderedTexture, bgl_TexCoord.xy) * brightness;
}
"""


class Brightness(ULFilter):

    def __init__(self, brightness=1.0, idx: int = None) -> None:
        self.settings = {'brightness': brightness}
        super().__init__(glsl, idx, {'brightness': self.settings})

    @property
    def brightness(self):
        return self.settings['brightness']

    @brightness.setter
    def brightness(self, val):
        self.settings['brightness'] = val



from bge import constraints
from bge import logic
from mathutils import Vector
from uplogic.nodes import ULActionNode
from uplogic.nodes import ULOutSocket
from uplogic.utils import is_invalid
from uplogic.utils import is_waiting
from uplogic.utils import not_met
import bpy


class ULSetCharacterWalkDir(ULActionNode):
    def __init__(self):
        ULActionNode.__init__(self)
        self.condition = None
        self.game_object = None
        self.walkDir = None
        self.local = False
        self.active = False
        self.done = None
        self.OUT = ULOutSocket(self, self.get_done)

    def get_done(self):
        return self.done

    def evaluate(self):
        self.done = False
        condition = self.get_input(self.condition)
        if not_met(condition):
            if self.active:
                game_object = self.get_input(self.game_object)
                physics = constraints.getCharacter(game_object)
                physics.walkDirection = Vector((0, 0, 0))
                self.active = False
            return
        elif not self.active:
            self.active = True
        game_object = self.get_input(self.game_object)
        local = self.local
        walkDir = self.get_input(self.walkDir)
        if is_waiting(game_object, local, walkDir):
            return
        self._set_ready()
        if is_invalid(game_object):
            return
        if local:
            walkDir = game_object.worldOrientation @ walkDir
        physics = constraints.getCharacter(game_object)
        physics.walkDirection = (
            walkDir /
            bpy.data.scenes[
                logic.getCurrentScene().name
            ].game_settings.physics_step_sub
        )
        self.done = True



from setuptools import setup
import os


def read_file(name):
    with open(os.path.join(os.path.dirname(__file__), name)) as f:
        return f.read()


version = 'v1.6.7'
shortdesc = "Uplogic utility for UPBGE."
longdesc = '\n\n'.join([read_file(name) for name in [
    'README.md',
    'LICENSE.md'
]])


setup(
    name='uplogic',
    version=version,
    description=shortdesc,
    long_description=longdesc,
    classifiers=[
        'License :: OSI Approved :: GNU General Public License v2 (GPLv2)',
        'Intended Audience :: Developers',
        'Topic :: Multimedia :: Graphics',
        'Programming Language :: Python',
    ],
    keywords='Blender UPBGE logic',
    author='Leopold Auersperg-Castell',
    author_email='lauersperg@gmx.at',
    url='https://github.com/UPBGE/uplogic',
    download_url='https://github.com/UPBGE/uplogic/archive/refs/tags/v1.6.7.tar.gz',
    license='GPLv2',
    packages=[
        'uplogic',
        'uplogic.animation',
        'uplogic.audio',
        'uplogic.data',
        'uplogic.events',
        'uplogic.input',
        'uplogic.logging',
        'uplogic.nodes',
        'uplogic.nodes.actions',
        'uplogic.nodes.conditions',
        'uplogic.nodes.parameters',
        'uplogic.physics',
        'uplogic.shaders',
        'uplogic.utils'
    ],
    zip_safe=True,
    install_requires=['setuptools']
)



"""
Author: Robert Niederreiter

Python based main loop for bge
==============================

Starting the custom mainloop
----------------------------

Add custom property ``__main__`` with value ``main.py`` to scene.

main.py contains:

.. code-block:: python

    from uplogic.mainloop import start

    start(max_fps=60, tick_idle=.001)

Using the mainloop
------------------

.. code-block:: python

    from uplogic.mainloop import MainLoop

    def on_start_callback():
        '''Gets called when custom mainloop starts.'''

    MainLoop.on_start(on_start_callback)

    def on_tick_callback():
        '''Gets called as fast as mainloop runs.'''

    MainLoop.on_tick(on_tick_callback)

    def on_update_callback():
        '''Gets called when enough time has elapsed to render the next frame.'''

    MainLoop.on_update(on_update_callback)

    def on_stop_callback():
        '''Gets called when custom mainloop is stopped.'''

    MainLoop.on_stop(on_stop_callback)


Stopping the mainloop
---------------------

By default, the mainloop gets stopped by SIGTERM. If you want to stop the
mainloop on purpose, use the following code:

.. code-block:: python

    from uplogic.mainloop import get_mainloop

    get_mainloop().stop(
"""

from collections import deque

from .utils import load_user_module
from .input import key_tap
import bge
import bpy
import signal
import time


class MainLoop:
    _handles = {}

    def __init__(self, max_fps=60, tick_idle=.001):
        self.max_fps = max_fps
        self.tick_idle = tick_idle
        self._last_frame = 0
        self._running = False
        self._evt_queue = deque()
        signal.signal(signal.SIGTERM, self._terminate)

    @property
    def max_fps(self):
        return self._max_fps

    @max_fps.setter
    def max_fps(self, value):
        self._max_fps = value
        self._frame_delay = 1. / value

    def start(self):
        self._running = True
        self._dispatch('start')
        try:
            while self._running:
                self._do_tick()
                time.sleep(self.tick_idle)
        except KeyboardInterrupt:
            self.stop()

    def stop(self):
        self._dispatch('stop')
        self._running = False
        bge.logic.endGame()

    def _terminate(self, *args):
        self.stop()

    def dispatch(self, event, *a, **kw):
        self._evt_queue.append((event, a, kw))

    def _dispatch_queue(self):
        while len(self._evt_queue) > 0:
            e, a, kw = self._evt_queue.popleft()
            self._dispatch(e, *a, **kw)

    def _dispatch(self, event, *a, **kw):
        for handle in self._handles.get(event, []):
            handle(*a, **kw)

    def _do_tick(self):
        now = time.time()
        self._dispatch('tick')
        fps = bge.logic.getAverageFrameRate()
        if self._last_frame + self._frame_delay <= now:# or fps < self.max_fps:
            self._last_frame = now
            self._dispatch('update')
            self._dispatch_queue()
            bge.logic.NextFrame()

    @classmethod
    def on(cls, **kw):
        for event, handle in kw.items():
            handles = cls._handles.setdefault(event, [])
            if handle not in handles:
                handles.append(handle)

    @classmethod
    def off(cls, event=None, handle=None):
        if event is not None and handle is None:
            if event in cls._handles:
                del cls._handles[event]
        elif event is None and handle is not None:
            for handles in cls._handles.values():
                if handle in handles:
                    handles.remove(handle)
        elif event is not None and handle is not None:
            handles = cls._handles.get(event, [])
            if handle in handles:
                handles.remove(handle)

    @classmethod
    def on_start(cls, handle):
        cls.on(start=handle)

    @classmethod
    def on_stop(cls, handle):
        cls.on(stop=handle)

    @classmethod
    def on_tick(cls, handle):
        cls.on(tick=handle)

    @classmethod
    def on_update(cls, handle):
        cls.on(update=handle)


_mainloop = None


def create_mainloop(max_fps, tick_idle):
    global _mainloop
    if _mainloop is not None:
        return _mainloop
    _mainloop = MainLoop(
        max_fps=max_fps,
        tick_idle=tick_idle
    )
    return _mainloop


def get_mainloop():
    global _mainloop
    return _mainloop


def start(max_fps=60, tick_idle=.001):
    """Start event loop.
    """
    eventloop = create_mainloop(max_fps, tick_idle)
    eventloop.start()


class ULLoop:

    def __init__(self, max_fps=-1, tick_idle=.00001) -> None:
        self.scene = bge.logic.getCurrentScene()
        logic_tree = bpy.data.scenes[self.scene.name].get('custom_mainloop_tree')
        if logic_tree:
            module = load_user_module(f'nl_{logic_tree.name.lower()}')
            logic_tree = module.get_tree(self.scene)
            MainLoop.on_update(logic_tree.network.evaluate)
        MainLoop.on_start(self.start)
        MainLoop.on_tick(self.tick)
        MainLoop.on_update(self.update)
        MainLoop.on_stop(self.stop)

        self.scene['uplogic.mainloop'] = self
        if max_fps == -1:
            max_fps = bpy.data.scenes[self.scene.name].game_settings.fps
        start(max_fps, tick_idle)

    def start(self):
        pass

    def tick(self):
        if key_tap('esc'):
            get_mainloop().stop()

    def update(self):
        pass

    def stop(self):
        pass



from bge import render
from bge.types import KX_GameObject as GameObject
from uplogic.nodes import ULActionNode, ULOutSocket
from uplogic.utils import is_invalid
from uplogic.utils import not_met
from uplogic.utils import rot_to
from uplogic.utils import move_to


class ULMoveToWithNavmesh(ULActionNode):

    class MotionPath(object):
        def __init__(self):
            self.points: list = []
            self.cursor: int = 0
            self.destination = None

        def next_point(self):
            if self.cursor < len(self.points):
                return self.points[self.cursor]
            else:
                return None

        def destination_changed(self, new_destination):
            return self.destination != new_destination

        def advance(self):
            self.cursor += 1
            return self.cursor < len(self.points)

    def __init__(self):
        ULActionNode.__init__(self)
        self.condition: bool = None
        self.moving_object: GameObject = None
        self.rotating_object: GameObject = None
        self.navmesh_object: GameObject = None
        self.destination_point = None
        self.move_dynamic: bool = None
        self.linear_speed: float = None
        self.reach_threshold: float = None
        self.look_at: bool = None
        self.rot_axis: int = None
        self.front_axis: int = None
        self.rot_speed: float = None
        self.visualize: bool = None
        self._motion_path = None
        self.done: bool = False
        self.finished: bool = False
        self.OUT = ULOutSocket(self.get_done)
        self.FINISHED = ULOutSocket(self.get_finished)
        self.POINT = ULOutSocket(self.get_point)

    def get_done(self):
        return self.done

    def get_finished(self):
        return self.finished

    def get_point(self):
        return self._motion_path.next_point()

    def evaluate(self):
        self.done = False
        condition: bool = self.get_input(self.condition)
        if not_met(condition):
            self._set_ready()
            return
        moving_object: GameObject = self.get_input(self.moving_object)
        rotating_object: GameObject = self.get_input(self.rotating_object)
        navmesh_object: GameObject = self.get_input(self.navmesh_object)
        destination_point = self.get_input(self.destination_point)
        move_dynamic: bool = self.get_input(self.move_dynamic)
        linear_speed: float = self.get_input(self.linear_speed)
        reach_threshold: float = self.get_input(self.reach_threshold)
        look_at: bool = self.get_input(self.look_at)
        rot_axis: int = self.get_input(self.rot_axis)
        front_axis: int = self.get_input(self.front_axis)
        rot_speed: float = self.get_input(self.rot_speed)
        visualize: bool = self.get_input(self.visualize)
        if is_invalid(
            destination_point,
            move_dynamic,
            linear_speed,
            reach_threshold,
            look_at,
            rot_axis,
            front_axis,
            rot_speed,
            visualize
        ):
            return
        if is_invalid(moving_object, navmesh_object):
            return
        if is_invalid(rotating_object):
            rotating_object = None
        self._set_ready()
        self.finished = False
        if (
            (self._motion_path is None) or
            (self._motion_path.destination_changed(destination_point))
        ):
            points = navmesh_object.findPath(
                moving_object.worldPosition,
                destination_point
            )
            motion_path = ULMoveToWithNavmesh.MotionPath()
            motion_path.points = points[1:]
            motion_path.destination = destination_point
            self._motion_path = motion_path
        next_point = self._motion_path.next_point()
        if visualize:
            points = [moving_object.worldPosition.copy()]
            points.extend(self._motion_path.points[self._motion_path.cursor:])
            points.append(self._motion_path.destination)
            for i, p in enumerate(points):
                if i < len(points) - 1:
                    render.drawLine(
                        p, points[i + 1], [1, 0, 0, 1]
                    )
        if next_point:
            tpf = self.network.time_per_frame
            if look_at and (rotating_object is not None):
                rot_to(
                    rot_axis,
                    rotating_object,
                    next_point,
                    front_axis,
                    rot_speed
                )
            ths = reach_threshold  # if next_point == self._motion_path.destination else .1  # noqa
            reached = move_to(
                moving_object,
                next_point,
                linear_speed,
                tpf,
                move_dynamic,
                ths
            )
            if reached:
                has_more = self._motion_path.advance()
                if not has_more:
                    self.finished = True
            self.done = True


from uplogic.nodes import ULParameterNode
from uplogic.utils import STATUS_READY


class ULGetOwner(ULParameterNode):
    def __init__(self):
        ULParameterNode.__init__(self)

    def setup(self, network):
        ULParameterNode.setup(self, network)
        self._set_status(STATUS_READY)
        self._set_value(network.get_owner())

    def reset(self):
        pass

    def evaluate(self):
        pass





from uplogic.nodes import ULOutSocket
from uplogic.nodes import ULParameterNode
from uplogic.utils import is_waiting


class ULInitEmptyList(ULParameterNode):
    def __init__(self):
        ULParameterNode.__init__(self)
        self.condition = None
        self.length = None
        self.items = None
        self.LIST = ULOutSocket(self, self.get_list)

    def get_list(self):
        length = self.get_input(self.length)
        if is_waiting(length):
            return
        return [None for x in range(length)]

    def evaluate(self):
        self._set_ready()



from uplogic.nodes import ULActionNode
from uplogic.nodes import ULOutSocket
from bge import events


class ULKeyLogger(ULActionNode):
    def __init__(self, pulse=False):
        ULActionNode.__init__(self)
        self.condition = None
        self.pulse = pulse
        self._key_logged = None
        self._key_code = None
        self._character = None
        self.KEY_LOGGED = ULOutSocket(self, self.get_key_logged)
        self.KEY_CODE = ULOutSocket(self, self.get_key_code)
        self.CHARACTER = ULOutSocket(self, self.get_character)

    def get_key_logged(self):
        return self._key_logged

    def get_key_code(self):
        return self._key_code

    def get_character(self):
        return self._character

    def reset(self):
        super().reset(self)
        self._key_logged = False
        self._key_code = None
        self._character = None

    def evaluate(self):
        self._set_ready()
        condition = self.get_input(self.condition)
        if not condition:
            return
        network = self.network
        keyboard_status = network.keyboard_events
        left_shift_status = keyboard_status[events.LEFTSHIFTKEY].active
        right_shift_status = keyboard_status[events.RIGHTSHIFTKEY].active
        shift_down = (
            left_shift_status or
            right_shift_status or
            network.capslock_pressed
        )
        active_events = network.active_keyboard_events
        active = (
            'active' if self.pulse
            else 'activated'
        )
        for keycode in active_events:
            event = active_events[keycode]
            if getattr(event, active):
                # something has been pressed
                self._character = events.EventToCharacter(
                    event.type,
                    shift_down
                )
                self._key_code = keycode
                self._key_logged = True
